import argparse
import numpy as np
import numpy as np
import torchvision.models.resnet as resnet
import struct
import os
import sys

def load_weights(dir):
    if not os.path.isdir(dir):
        print("Path %s is not a dir" % dir)
        sys.exit(1)

    print("-----------loading weights------------")
    resnet50 = resnet.resnet50(pretrained=True)
    resnet50.eval()
    dic = resnet50.state_dict()
    # numpy stores weights in output channel, input channel, h, w order
    # we want to transpose to output channel, h, w, input channel order
    for k in dic.keys():
        weight = dic[k].cpu().detach().numpy()
        weight = weight.astype(np.float32)
        print("weight shape before transpose")
        print("%s,%s,%d" % (k, str(weight.shape), len(weight.shape)))
        if len(weight.shape) == 4:
            weight = np.transpose(weight, (1, 2, 3, 0))
        if len(weight.shape) == 2:
            weight = np.transpose(weight, (1, 0))
        print("weight shape after transpose")
        print("%s,%s,%d" % (k, str(weight.shape), len(weight.shape)))

        data = weight.tobytes()
        path = os.path.join(dir, k.replace('.','_'))
        with open(path + ".data", "wb") as f:
            f.write(data)
        with open(path + '_shape.data', 'wb') as f:
            f.write(struct.pack('i', len(weight.shape)))
            for i in list(reversed(range(len(weight.shape)))):
                f.write(struct.pack('i', weight.shape[i]))

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: load_weights destdir")
        sys.exit(1)
    load_weights(sys.argv[1])



#!/usr/bin/python3

# Halide tutorial lesson 6.

# This lesson demonstrates how to evaluate a hl.Func over a domain that
# does not start at (0, 0).

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_06_realizing_over_shifted_domains
# in a shell with the current directory at python_bindings/

import halide as hl


def main():

    # The last lesson was quite involved, and scheduling complex
    # multi-stage pipelines is ahead of us. As an interlude, let's
    # consider something easy: evaluating funcs over rectangular
    # domains that do not start at the origin.

    # We define our familiar gradient function.
    gradient = hl.Func("gradient")
    x, y = hl.Var("x"), hl.Var("y")
    gradient[x, y] = x + y

    # And turn on tracing so we can see how it is being evaluated.
    gradient.trace_stores()

    # Previously we've realized gradient like so:
    #
    # gradient.realize([8, 8])
    #
    # This does three things internally:
    # 1) Generates code than can evaluate gradient over an arbitrary
    # rectangle.
    # 2) Allocates a new 8 x 8 image.
    # 3) Runs the generated code to evaluate gradient for all x, y
    # from (0, 0) to (7, 7) and puts the result into the image.
    # 4) Returns the new image as the result of the realize call.

    # What if we're managing memory carefully and don't want Halide
    # to allocate a new image for us? We can call realize another
    # way. We can pass it an image we would like it to fill in. The
    # following evaluates our hl.Func into an existing image:
    print("Evaluating gradient from (0, 0) to (7, 7)")
    result = hl.Buffer(hl.Int(32), [8, 8])
    gradient.realize(result)

    # Let's check it did what we expect:
    for yy in range(8):
        for xx in range(8):
            assert result[xx, yy] == xx + yy, "Something went wrong!"

    # Now let's evaluate gradient over a 5 x 7 rectangle that starts
    # somewhere else -- at position (100, 50). So x and y will run
    # from (100, 50) to (104, 56) inclusive.

    # We start by creating an image that represents that rectangle:
    # In the constructor we tell it the size.
    shifted = hl.Buffer(hl.Int(32), [5, 7])
    shifted.set_min([100, 50])  # Then we tell it the top-left corner.

    print("Evaluating gradient from (100, 50) to (104, 56)")

    # Note that this won't need to compile any new code, because when
    # we realized it the first time, we generated code capable of
    # evaluating gradient over an arbitrary rectangle.
    gradient.realize(shifted)

    # From C++, we also access the image object using coordinates
    # that start at (100, 50).
    for yy in range(50, 57):
        for xx in range(100, 105):
            assert shifted[xx, yy] == xx + yy, "Something went wrong!"

    # The image 'shifted' stores the value of our hl.Func over a domain
    # that starts at (100, 50), so asking for shifted(0, 0) would in
    # fact read out-of-bounds and probably crash.

    # What if we want to evaluate our hl.Func over some region that
    # isn't rectangular? Too bad. Halide only does rectangles :)

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/python3

# Halide tutorial lesson 14: The Halide type system

# This lesson more precisely describes Halide's type system.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_14_types
# in a shell with the current directory at python_bindings/


import halide as hl


def main():

    # All Exprs have a scalar type, and all Funcs evaluate to one or
    # more scalar types. The scalar types in Halide are unsigned
    # integers of various bit widths, signed integers of the same set
    # of bit widths, floating point numbers in single and double
    # precision, and opaque handles (equivalent to void *). The
    # following array contains all the legal types.

    valid_halide_types = [
        hl.UInt(8), hl.UInt(16), hl.UInt(32), hl.UInt(64),
        hl.Int(8), hl.Int(16), hl.Int(32), hl.Int(64),
        hl.Float(32), hl.Float(64), hl.Handle()]

    # Constructing and inspecting types.
    if True:
        # You can programmatically examine the properties of a Halide
        # type. This is useful when you write a C++ function that has
        # hl.Expr arguments and you wish to check their types:
        assert hl.UInt(8).bits() == 8
        assert hl.Int(8).is_int()

        # You can also programmatically construct Types as a function of other
        # Types.
        t = hl.UInt(8)
        t = t.with_bits(t.bits() * 2)
        assert t == hl.UInt(16)

        # The Type struct is also capable of representing vector types,
        # but this is reserved for Halide's internal use. You should
        # vectorize code by using hl.Func::vectorize, not by attempting to
        # construct vector expressions directly. You may encounter vector
        # types if you programmatically manipulate lowered Halide code,
        # but this is an advanced topic (see
        # hl.Func::add_custom_lowering_pass).

        # You can query any Halide hl.Expr for its type. An hl.Expr
        # representing a hl.Var has type hl.Int(32):
        x = hl.Var("x")
        assert hl.Expr(x).type() == hl.Int(32)

        # Most transcendental functions in Halide hl.cast their inputs to a
        # hl.Float(32) and return a hl.Float(32):
        assert hl.sin(x).type() == hl.Float(32)

        # You can hl.cast an hl.Expr from one Type to another using the hl.cast
        # operator:
        assert hl.cast(hl.UInt(8), x).type() == hl.UInt(8)

        # You can also query any defined hl.Func for the types it produces.
        f1 = hl.Func("f1")
        f1[x] = hl.cast(hl.UInt(8), x)
        assert f1.types()[0] == hl.UInt(8)

        f2 = hl.Func("f2")
        f2[x] = (x, hl.sin(x))
        assert f2.types()[0] == hl.Int(32) and f2.types()[1] == hl.Float(32)

    # Type promotion rules.
    if True:
        # When you combine Exprs of different types (e.g. using '+',
        # '*', etc), Halide uses a system of type promotion
        # rules. These differ to C's rules. To demonstrate these
        # we'll make some Exprs of each type.
        x = hl.Var("x")
        u8 = hl.cast(hl.UInt(8), x)
        u16 = hl.cast(hl.UInt(16), x)
        u32 = hl.cast(hl.UInt(32), x)
        u64 = hl.cast(hl.UInt(64), x)
        s8 = hl.cast(hl.Int(8), x)
        s16 = hl.cast(hl.Int(16), x)
        s32 = hl.cast(hl.Int(32), x)
        s64 = hl.cast(hl.Int(64), x)
        f32 = hl.cast(hl.Float(32), x)
        f64 = hl.cast(hl.Float(64), x)

        # The rules are as follows, and are applied in the order they are
        # written below.

        # 1) It is an error to hl.cast or use arithmetic operators on Exprs of
        # type hl.Handle().

        # 2) If the types are the same, then no type conversions occur.
        for t in valid_halide_types:
            # Skip the handle type.
            if t.is_handle():
                continue
            e = hl.cast(t, x)
            assert (e + e).type() == e.type()

        # 3) If one type is a float but the other is not, then the
        # non-float argument is promoted to a float (possibly causing a
        # loss of precision for large integers).
        assert (u8 + f32).type() == hl.Float(32)
        assert (f32 + s64).type() == hl.Float(32)
        assert (u16 + f64).type() == hl.Float(64)
        assert (f64 + s32).type() == hl.Float(64)

        # 4) If both types are float, then the narrower argument is
        # promoted to the wider bit-width.
        assert (f64 + f32).type() == hl.Float(64)

        # The rules above handle all the floating-point cases. The
        # following three rules handle the integer cases.

        # 5) If one of the expressions is an integer constant, then it is
        # coerced to the type of the other expression.
        assert (u32 + 3).type() == hl.UInt(32)
        assert (3 + s16).type() == hl.Int(16)

        # If this rule would cause the integer to overflow, then Halide
        # will trigger an error, e.g. uncommenting the following line
        # will cause this program to terminate with an error.
        # hl.Expr bad = u8 + 257

        # 6) If both types are unsigned integers, or both types are
        # signed integers, then the narrower argument is promoted to
        # wider type.
        assert (u32 + u8).type() == hl.UInt(32)
        assert (s16 + s64).type() == hl.Int(64)

        # 7) If one type is signed and the other is unsigned, both
        # arguments are promoted to a signed integer with the greater of
        # the two bit widths.
        assert (u8 + s32).type() == hl.Int(32)
        assert (u32 + s8).type() == hl.Int(32)

        # Note that this may silently overflow the unsigned type in the
        # case where the bit widths are the same.
        assert (u32 + s32).type() == hl.Int(32)

        if False:  # evaluate<X> not yet exposed to python
            # When an unsigned hl.Expr is converted to a wider signed type in
            # this way, it is first widened to a wider unsigned type
            # (zero-extended), and then reinterpreted as a signed
            # integer. I.e. casting the hl.UInt(8) value 255 to an hl.Int(32)
            # produces 255, not -1.
            # int32_t result32 =
            # evaluate<int>(hl.cast<int32_t>(hl.cast<uint8_t>(255)))
            assert result32 == 255

            # When a signed type is explicitly converted to a wider unsigned
            # type with the hl.cast operator (the type promotion rules will
            # never do this automatically), it is first converted to the
            # wider signed type (sign-extended), and then reinterpreted as
            # an unsigned integer. I.e. casting the hl.Int(8) value -1 to a
            # hl.UInt(16) produces 65535, not 255.
            # uint16_t result16 =
            # evaluate<uint16_t>(hl.cast<uint16_t>(hl.cast<int8_t>(-1)))
            assert result16 == 65535

    # The type hl.Handle().
    if True:
        # hl.Handle is used to represent opaque pointers. Applying
        # type_of to any pointer type will return hl.Handle()

        # Handles are always stored as 64-bit, regardless of the compilation
        # target.
        assert hl.Handle().bits() == 64

        # The main use of an hl.Expr of type hl.Handle is to pass
        # it through Halide to other external code.

    # Generic code.
    if True:
        # The main explicit use of Type in Halide is to write Halide
        # code parameterized by a Type. In C++ you'd do this with
        # templates. In Halide there's no need - you can inspect and
        # modify the types dynamically at C++ runtime instead. The
        # function defined below averages two expressions of any
        # equal numeric type.
        x = hl.Var("x")
        assert average(hl.cast(hl.Float(32), x), 3.0).type() == hl.Float(32)
        assert average(x, 3).type() == hl.Int(32)
        assert average(hl.cast(hl.UInt(8), x), hl.cast(hl.UInt(8), 3)).type() == hl.UInt(8)

    print("Success!")

    return 0


def average(a, b):

    if type(a) is not hl.Expr:
        a = hl.Expr(a)

    if type(b) is not hl.Expr:
        b = hl.Expr(b)

    "hl.Expr average(hl.Expr a, hl.Expr b)"
    # Types must match.
    assert a.type() == b.type()

    # For floating point types:
    if (a.type().is_float()):
        # The '2' will be promoted to the floating point type due to
        # rule 3 above.
        return (a + b) / 2

    # For integer types, we must compute the intermediate value in a
    # wider type to avoid overflow.
    narrow = a.type()
    wider = narrow.with_bits(narrow.bits() * 2)
    a = hl.cast(wider, a)
    b = hl.cast(wider, b)
    return hl.cast(narrow, (a + b) / 2)


if __name__ == "__main__":
    main()



import unittest
from model import Model
from onnx import helper
from onnx import TensorProto
import numpy as np


class ModelTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_empty_model(self):
        model = Model()
        with self.assertRaises(Exception):
            model.GenerateSchedule()
        with self.assertRaises(Exception):
            model.PrintLoopNest()
        with self.assertRaises(Exception):
            model.PrintLoweredStatement()

    def test_small_model(self):
        # Create one input
        X = helper.make_tensor_value_info('IN', TensorProto.FLOAT, [2, 3])
        # Create one output
        Y = helper.make_tensor_value_info('OUT', TensorProto.FLOAT, [2, 3])
        # Create a node
        node_def = helper.make_node('Abs', ['IN'], ['OUT'])

        # Create the model
        graph_def = helper.make_graph([node_def], "test-model", [X], [Y])
        onnx_model = helper.make_model(graph_def,
                                       producer_name='onnx-example')

        model = Model()
        model.BuildFromOnnxModel(onnx_model)
        schedule = model.OptimizeSchedule()
        schedule = schedule.replace('\n', ' ')
        expected_schedule = r'.*Func OUT = pipeline.get_func\(1\);.+'
        self.assertRegex(schedule, expected_schedule)

        input = (np.random.rand(2, 3) - 0.5).astype('float32')
        outputs = model.run([input])
        self.assertEqual(1, len(outputs))
        output = outputs[0]
        expected = np.abs(input)
        np.testing.assert_allclose(expected, output)

    def test_scalars(self):
        # Create 2 inputs
        X = helper.make_tensor_value_info('A', TensorProto.INT32, [])
        Y = helper.make_tensor_value_info('B', TensorProto.INT32, [])
        # Create one output
        Z = helper.make_tensor_value_info('C', TensorProto.INT32, [])
        # Create a node
        node_def = helper.make_node('Add', ['A', 'B'], ['C'])

        # Create the model
        graph_def = helper.make_graph([node_def], "scalar-model", [X, Y], [Z])
        onnx_model = helper.make_model(graph_def,
                                       producer_name='onnx-example')

        model = Model()
        model.BuildFromOnnxModel(onnx_model)
        schedule = model.OptimizeSchedule()
        schedule = schedule.replace('\n', ' ')        
        expected_schedule = r'.*Func C = pipeline.get_func\(2\);.+'
        self.assertRegex(schedule, expected_schedule)

        input1 = np.random.randint(-10, 10, size=()).astype('int32')
        input2 = np.random.randint(-10, 10, size=()).astype('int32')
        outputs = model.run([input1, input2])
        self.assertEqual(1, len(outputs))
        output = outputs[0]
        expected = input1 + input2
        np.testing.assert_allclose(expected, output)

    def test_model_with_initializer(self):
        X = helper.make_tensor_value_info('X', TensorProto.FLOAT, [3, 1])
        Z2 = helper.make_tensor_value_info('Z2', TensorProto.FLOAT, [2, 3, 6])

        expand_node_def = helper.make_node('Expand', ['X', 'Y'], ['Z1'])
        cast_node_def = helper.make_node('Scale', ['Z1'], ['Z2'])

        graph_def = helper.make_graph([expand_node_def, cast_node_def],
            "test-node",
            [X],
            [Z2],
            initializer=[
                helper.make_tensor('Y', TensorProto.INT64, (3,), (2, 1, 6))])
        onnx_model = helper.make_model(graph_def,
                                       producer_name='onnx-example')
        model = Model()
        model.BuildFromOnnxModel(onnx_model)
        input_data = np.random.rand(3, 1).astype(np.float32)
        outputs = model.run([input_data])
        expected = input_data * np.ones([2, 1, 6], dtype=np.float32)
        np.testing.assert_allclose(expected, outputs[0])

    def test_tensors_rank_zero(self):
        X = helper.make_tensor_value_info('X', TensorProto.FLOAT, [3, 2])
        S1 = helper.make_tensor_value_info('S1', TensorProto.INT64, [])
        S2 = helper.make_tensor_value_info('S2', TensorProto.FLOAT, [])

        size_node = helper.make_node('Size', ['X'], ['S1'])

        graph_def = helper.make_graph([size_node],
            "rank_zero_test",
            [X],
            [S1, S2],
            initializer=[
                helper.make_tensor('S2', TensorProto.FLOAT, (), (3.14,))])
        onnx_model = helper.make_model(graph_def,
                                       producer_name='onnx-example')
        model = Model()
        model.BuildFromOnnxModel(onnx_model)
        input_data = np.random.rand(3, 2).astype(np.float32)
        outputs = model.run([input_data])
        self.assertEqual(6, outputs[0])
        self.assertAlmostEqual(3.14, outputs[1])




import unittest
import onnx.backend.test
import halide_as_onnx_backend as halide_backend

# This is a pytest magic variable to load extra plugins
#pytest_plugins = 'onnx.backend.test.report',

backend_test = onnx.backend.test.BackendTest(halide_backend, __name__)

exclude_test_patterns = (
    r'(test_hardsigmoid'  # Does not support Hardsigmoid.
     '|test_hardmax'  # Does not support Hardmax.
     '|test_depthtospace.*'  # Does not support DepthToSpace.
     '|test_.*pool_with_argmax.*'  # argmax not yet supported
     '|test_.*pool.*same.*'  # same padding not yet supported
     '|test_.*unpool.*'  # unpool not yet supported
     '|test_.*convtranspose.*'   # Not supported yet
     '|test_.*ConvTranspose.*'   # Not supported yet
     '|test_.*Conv.*_dilated.*'  # Not supported yet
     '|test_maxpool.*dilation.*'  # MaxPool doesn't support dilation yet
     '|test_.*mvn.*'  # MeanVarianceNormalization is not supported.
     '|test_.*pool.*ceil.*'  # ceil mode is not supported yet
     '|test_.*like.*'  # Needs implementation
     '|test_.*instancenorm.*'  # not supported yet
     '|test_reshape.*'
     '|test_shrink.*'
     '|test_tile.*'
     '|test_dynamic_slice.*'  # not supported yet
     '|test_arg.*'  # not supported yet
     '|test_top.*'  # not supported yet
     '|test_resize.*'  # opset 10 is not supported yet
     '|test_compress_.*'  # not supported yet
     '|test_nonzero.*'  # not supported yet
     '|test_tfidfvectorizer.*'  # not supported yet
     '|test_cast_FLOAT_to_STRING.*'  # not supported yet
     '|test_cast_STRING_to_FLOAT.*'  # not supported yet
     '|test_.*scatter.*'  # not supported yet
     '|test_.*upsample.*'  # not supported yet
     '|test_.*qlinear.*'  # skip quantized op test
     '|test_.*quantize.*'  # skip quantized op test
     '|test_.*matmulinteger.*'  # skip quantized op test
     '|test_.*convinteger.*'  # skip quantized op test
     '|test_mod_.*'  # not supported yet
     '|test_cumsum_.*'  # not supported yet
     '|test_bitshift_.*'  # not supported yet
     '|test_nonmaxsuppression.*'  # not supported yet
     '|test_reversesequence.*'  # not supported yet
     '|test_roialign.*'  # not supported yet
     '|test_round.*'  # not supported yet
     '|test_scan.*'  # Scan not supported yet
     '|test_strnorm.*'  # not supported yet
     '|test_.*index.*'  # Indexing not supported
     '|test_.*chunk.*'  # chunk not supported
     '|test_operator_symbolic_override.*'  # InstanceNormalization not supported yet
     '|test_.*FLOAT16*'  # FLOAT16 not supported yet
     '|test_densenet.*'   # Times out
     '|test_inception_v2.*'  # Padding type not supported for pooling
     ')'
)


backend_test.exclude(exclude_test_patterns)

# import all test cases at global scope to make them visible to python.unittest
globals().update(backend_test
                 .enable_report()
                 .test_cases)


if __name__ == '__main__':
    unittest.main()



import halide as hl

x = hl.Var("x")
y = hl.Var("y")
z = hl.Var("z")


@hl.alias(
    addconstantpy_with_offset_42={"extra_int": 42},
    addconstantpy_with_negative_offset={"extra_int": -1},
)
@hl.generator(name="addconstantpy")
class AddConstantGenerator:
    extra_int = hl.GeneratorParam(0)

    scalar_uint1 = hl.InputScalar(hl.Bool())
    scalar_uint8 = hl.InputScalar(hl.UInt(8))
    scalar_uint16 = hl.InputScalar(hl.UInt(16))
    scalar_uint32 = hl.InputScalar(hl.UInt(32))
    scalar_uint64 = hl.InputScalar(hl.UInt(64))
    scalar_int8 = hl.InputScalar(hl.Int(8))
    scalar_int16 = hl.InputScalar(hl.Int(16))
    scalar_int32 = hl.InputScalar(hl.Int(32))
    scalar_int64 = hl.InputScalar(hl.Int(64))
    scalar_float = hl.InputScalar(hl.Float(32))
    scalar_double = hl.InputScalar(hl.Float(64))

    input_uint8 = hl.InputBuffer(hl.UInt(8), 1)
    input_uint16 = hl.InputBuffer(hl.UInt(16), 1)
    input_uint32 = hl.InputBuffer(hl.UInt(32), 1)
    input_uint64 = hl.InputBuffer(hl.UInt(64), 1)
    input_int8 = hl.InputBuffer(hl.Int(8), 1)
    input_int16 = hl.InputBuffer(hl.Int(16), 1)
    input_int32 = hl.InputBuffer(hl.Int(32), 1)
    input_int64 = hl.InputBuffer(hl.Int(64), 1)
    input_float = hl.InputBuffer(hl.Float(32), 1)
    input_double = hl.InputBuffer(hl.Float(64), 1)
    input_half = hl.InputBuffer(hl.Float(16), 1)
    input_2d = hl.InputBuffer(hl.Int(8), 2)
    input_3d = hl.InputBuffer(hl.Int(8), 3)

    output_uint8 = hl.OutputBuffer(hl.UInt(8), 1)
    output_uint16 = hl.OutputBuffer(hl.UInt(16), 1)
    output_uint32 = hl.OutputBuffer(hl.UInt(32), 1)
    output_uint64 = hl.OutputBuffer(hl.UInt(64), 1)
    output_int8 = hl.OutputBuffer(hl.Int(8), 1)
    output_int16 = hl.OutputBuffer(hl.Int(16), 1)
    output_int32 = hl.OutputBuffer(hl.Int(32), 1)
    output_int64 = hl.OutputBuffer(hl.Int(64), 1)
    output_float = hl.OutputBuffer(hl.Float(32), 1)
    output_double = hl.OutputBuffer(hl.Float(64), 1)
    output_half = hl.OutputBuffer(hl.Float(16), 1)
    output_2d = hl.OutputBuffer(hl.Int(8), 2)
    output_3d = hl.OutputBuffer(hl.Int(8), 3)

    def generate(self):
        g = self
        g.add_requirement(g.scalar_int32 != 0)  # error_args omitted for this case
        g.add_requirement(g.scalar_int32 > 0, "negative values are bad", g.scalar_int32)

        g.output_uint8[x] = g.input_uint8[x] + g.scalar_uint8
        g.output_uint16[x] = g.input_uint16[x] + g.scalar_uint16
        g.output_uint32[x] = g.input_uint32[x] + g.scalar_uint32
        g.output_uint64[x] = g.input_uint64[x] + g.scalar_uint64
        g.output_int8[x] = g.input_int8[x] + g.scalar_int8
        g.output_int16[x] = g.input_int16[x] + g.scalar_int16
        g.output_int32[x] = g.input_int32[x] + g.scalar_int32
        g.output_int64[x] = g.input_int64[x] + g.scalar_int64
        g.output_float[x] = g.input_float[x] + g.scalar_float
        g.output_double[x] = g.input_double[x] + g.scalar_double
        g.output_half[x] = g.input_half[x] + hl.cast(hl.Float(16), g.scalar_float)
        g.output_2d[x, y] = g.input_2d[x, y] + g.scalar_int8
        g.output_3d[x, y, z] = g.input_3d[x, y, z] + g.scalar_int8 + g.extra_int


if __name__ == "__main__":
    hl.main()



#!/usr/bin/python3

# Halide tutorial lesson 8

# This lesson demonstrates how schedule multi-stage pipelines.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_08_scheduling_2
# in a shell with the current directory at python_bindings/

import halide as hl
import numpy as np
import math


def main():
    # First we'll declare some Vars to use below.
    x, y = hl.Var("x"), hl.Var("y")

    # Let's examine various scheduling options for a simple two stage
    # pipeline. We'll start with the default schedule:
    if True:
        print("=" * 50)
        producer, consumer = hl.Func(
            "producer_default"), hl.Func("consumer_default")

        # The first stage will be some simple pointwise math similar
        # to our familiar gradient function. The value at position x,
        # y is the sqrt of product of x and y.
        producer[x, y] = hl.sqrt(x * y)

        # Now we'll add a second stage which adds together multiple
        # points in the first stage.
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # We'll turn on tracing for both functions.
        consumer.trace_stores()
        producer.trace_stores()

        # And evaluate it over a 5x5 box.
        print("\nEvaluating producer-consumer pipeline with default schedule")
        consumer.realize([4, 4])

        # There were no messages about computing values of the
        # producer. This is because the default schedule fully
        # inlines 'producer' into 'consumer'. It is as if we had
        # written the following code instead:

        # consumer[x, y] = (sqrt(x * y) +
        #                   sqrt(x * (y + 1)) +
        #                   sqrt((x + 1) * y) +
        #                   sqrt((x + 1) * (y + 1)))

        # All calls to 'producer' have been replaced with the body of
        # 'producer', with the arguments subtituted in for the
        # variables.

        # The equivalent C code is:
        result = np.empty((4, 4), dtype=np.float32)
        for yy in range(4):
            for xx in range(4):
                result[yy][xx] = (math.sqrt(xx * yy) +
                                  math.sqrt(xx * (yy + 1)) +
                                  math.sqrt((xx + 1) * yy) +
                                  math.sqrt((xx + 1) * (yy + 1)))

        print()

        # If we look at the loop nest, the producer doesn't appear
        # at all. It has been inlined into the consumer.
        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

    # Next we'll examine the next simplest option - computing all
    # values required in the producer before computing any of the
    # consumer. We call this schedule "root".
    if True:
        print("=" * 50)
        # Start with the same function definitions:
        producer, consumer = hl.Func("producer_root"), hl.Func("consumer_root")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Tell Halide to evaluate all of producer before any of consumer.
        producer.compute_root()

        # Turn on tracing.
        consumer.trace_stores()
        producer.trace_stores()

        # Compile and run.
        print("\nEvaluating producer.compute_root()")
        consumer.realize([4, 4])

        # Reading the output we can see that:
        # A) There were stores to producer.
        # B) They all happened before any stores to consumer.

        # Equivalent C:
        result = np.empty((4, 4), dtype=np.float32)

        # Allocate some temporary storage for the producer.
        producer_storage = np.empty((5, 5), dtype=np.float32)

        # Compute the producer.
        for yy in range(5):
            for xx in range(5):
                producer_storage[yy][xx] = math.sqrt(xx * yy)

        # Compute the consumer. Skip the prints this time.
        for yy in range(4):
            for xx in range(4):
                result[yy][xx] = (producer_storage[yy][xx] +
                                  producer_storage[yy + 1][xx] +
                                  producer_storage[yy][xx + 1] +
                                  producer_storage[yy + 1][xx + 1])

        # Note that consumer was evaluated over a 4x4 box, so Halide
        # automatically inferred that producer was needed over a 5x5
        # box. This is the same 'bounds inference' logic we saw in
        # the previous lesson, where it was used to detect and avoid
        # out-of-bounds reads from an input image.

        # If we print the loop nest, we'll see something very
        # similar to the C above.
        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

    # Let's compare the two approaches above from a performance
    # perspective.

    # Full inlining (the default schedule):
    # - Temporary memory allocated: 0
    # - Loads: 0
    # - Stores: 16
    # - Calls to sqrt: 64

    # producer.compute_root():
    # - Temporary memory allocated: 25 floats
    # - Loads: 64
    # - Stores: 39
    # - Calls to sqrt: 25

    # There's a trade-off here. Full inlining used minimal temporary
    # memory and memory bandwidth, but did a whole bunch of redundant
    # expensive math (calling sqrt). It evaluated most points in
    # 'producer' four times. The second schedule,
    # producer.compute_root(), did the mimimum number of calls to
    # sqrt, but used more temporary memory and more memory bandwidth.

    # In any given situation the correct choice can be difficult to
    # make. If you're memory-bandwidth limited, or don't have much
    # memory (e.g. because you're running on an old cell-phone), then
    # it can make sense to do redundant math. On the other hand, sqrt
    # is expensive, so if you're compute-limited then fewer calls to
    # sqrt will make your program faster. Adding vectorization or
    # multi-core parallelism tilts the scales in favor of doing
    # redundant work, because firing up multiple cpu cores increases
    # the amount of math you can do per second, but doesn't increase
    # your system memory bandwidth or capacity.

    # We can make choices in between full inlining and
    # compute_root. Next we'll alternate between computing the
    # producer and consumer on a per-scanline basis:
    if True:
        print("=" * 50)
        # Start with the same function definitions:
        producer, consumer = hl.Func("producer_y"), hl.Func("consumer_y")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Tell Halide to evaluate producer as needed per y coordinate
        # of the consumer:
        producer.compute_at(consumer, y)

        # This places the code that computes the producer just
        # *inside* the consumer's for loop over y, as in the
        # equivalent C below.

        # Turn on tracing.
        producer.trace_stores()
        consumer.trace_stores()

        # Compile and run.
        print("\nEvaluating producer.compute_at(consumer, y)")
        consumer.realize([4, 4])

        # Reading the log you should see that producer and consumer
        # alternate on a per-scanline basis. Let's look at the
        # equivalent C:
        result = np.empty((4, 4), dtype=np.float32)

        # There's an outer loop over scanlines of consumer:
        for yy in range(4):

            # Allocate space and compute enough of the producer to
            # satisfy this single scanline of the consumer. This
            # means a 5x2 box of the producer.
            producer_storage = np.empty((2, 5), dtype=np.float32)
            for py in range(yy, yy + 2):
                for px in range(5):
                    producer_storage[py - yy][px] = math.sqrt(px * py)

            # Compute a scanline of the consumer.
            for xx in range(4):
                result[yy][xx] = (producer_storage[0][xx] +
                                  producer_storage[1][xx] +
                                  producer_storage[0][xx + 1] +
                                  producer_storage[1][xx + 1])

        # Again, if we print the loop nest, we'll see something very
        # similar to the C above.
        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

        # The performance characteristics of this strategy are in
        # between inlining and compute root. We still allocate some
        # temporary memory, but less that compute_root, and with
        # better locality (we load from it soon after writing to it,
        # so for larger images, values should still be in cache). We
        # still do some redundant work, but less than full inlining:

        # producer.compute_at(consumer, y):
        # - Temporary memory allocated: 10 floats
        # - Loads: 64
        # - Stores: 56
        # - Calls to sqrt: 40

    # We could also say producer.compute_at(consumer, x), but this
    # would be very similar to full inlining (the default
    # schedule). Instead let's distinguish between the loop level at
    # which we allocate storage for producer, and the loop level at
    # which we actually compute it. This unlocks a few optimizations.
    if True:
        print("=" * 50)
        producer = hl.Func("producer_store_root_compute_y")
        consumer = hl.Func("consumer_store_root_compute_y")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Tell Halide to make a buffer to store all of producer at
        # the outermost level:
        producer.store_root()
        # ... but compute it as needed per y coordinate of the
        # consumer.
        producer.compute_at(consumer, y)

        producer.trace_stores()
        consumer.trace_stores()

        print("\nEvaluating producer.store_root().compute_at(consumer, y)")
        consumer.realize([4, 4])

        # Reading the log you should see that producer and consumer
        # again alternate on a per-scanline basis. It computes a 5x2
        # box of the producer to satisfy the first scanline of the
        # consumer, but after that it only computes a 5x1 box of the
        # output for each new scanline of the consumer!
        #
        # Halide has detected that for all scanlines except for the
        # first, it can reuse the values already sitting in the
        # buffer we've allocated for producer. Let's look at the
        # equivalent C:

        result = np.empty((4, 4), dtype=np.float32)

        # producer.store_root() implies that storage goes here:
        producer_storage = np.empty((5, 5), dtype=np.float32)

        # There's an outer loop over scanlines of consumer:
        for yy in range(4):

            # Compute enough of the producer to satisfy this scanline
            # of the consumer.
            for py in range(yy, yy + 2):

                # Skip over rows of producer that we've already
                # computed in a previous iteration.
                if yy > 0 and py == yy:
                    continue

                for px in range(5):
                    producer_storage[py][px] = math.sqrt(px * py)

            # Compute a scanline of the consumer.
            for xx in range(4):
                result[yy][xx] = (producer_storage[yy][xx] +
                                  producer_storage[yy + 1][xx] +
                                  producer_storage[yy][xx + 1] +
                                  producer_storage[yy + 1][xx + 1])

        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

        # The performance characteristics of this strategy are pretty
        # good! The numbers are similar compute_root, except locality
        # is better. We're doing the minimum number of sqrt calls,
        # and we load values soon after they are stored, so we're
        # probably making good use of the cache:

        # producer.store_root().compute_at(consumer, y):
        # - Temporary memory allocated: 10 floats
        # - Loads: 64
        # - Stores: 39
        # - Calls to sqrt: 25

        # Note that my claimed amount of memory allocated doesn't
        # match the reference C code. Halide is performing one more
        # optimization under the hood. It folds the storage for the
        # producer down into a circular buffer of two
        # scanlines. Equivalent C would actually look like this:

        if True:
            # Actually store 2 scanlines instead of 5
            producer_storage = np.empty((2, 5), dtype=np.float32)

            for yy in range(4):
                for py in range(yy, yy + 2):

                    if yy > 0 and py == yy:
                        continue

                    for px in range(5):
                        # Stores to producer_storage have their y coordinate
                        # bit-masked.
                        producer_storage[py & 1][px] = math.sqrt(px * py)

                # Compute a scanline of the consumer.
                for xx in range(4):
                    # Loads from producer_storage have their y coordinate
                    # bit-masked.
                    result[yy][xx] = (producer_storage[yy & 1][xx] +
                                      producer_storage[(yy + 1) & 1][xx] +
                                      producer_storage[yy & 1][xx + 1] +
                                      producer_storage[(yy + 1) & 1][xx + 1])

    # We can do even better, by leaving the storage outermost, but
    # moving the computation into the innermost loop:
    if True:
        print("=" * 50)
        producer = hl.Func("producer_store_root_compute_x")
        consumer = hl.Func("consumer_store_root_compute_x")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Store outermost, compute innermost.
        producer.store_root().compute_at(consumer, x)

        producer.trace_stores()
        consumer.trace_stores()

        print("\nEvaluating producer.store_root().compute_at(consumer, x)")
        consumer.realize([4, 4])

        # Reading the log, you should see that producer and consumer
        # now alternate on a per-pixel basis. Here's the equivalent C:

        result = np.empty((4, 4), dtype=np.float32)

        # producer.store_root() implies that storage goes here, but
        # we can fold it down into a circular buffer of two
        # scanlines:
        producer_storage = np.empty((2, 5), dtype=np.float32)

        # For every pixel of the consumer:
        for yy in range(4):
            for xx in range(4):

                # Compute enough of the producer to satisfyy this
                # pixxel of the consumer, but skip values that we've
                # alreadyy computed:
                if (yy == 0) and (xx == 0):
                    producer_storage[yy & 1][xx] = math.sqrt(xx * yy)
                if yy == 0:
                    producer_storage[yy & 1][xx + 1] = math.sqrt((xx + 1) * yy)
                if xx == 0:
                    producer_storage[(yy + 1) & 1][xx] = math.sqrt(xx * (yy + 1))

                producer_storage[(yy + 1) & 1][xx + 1] = math.sqrt((xx + 1) * (yy + 1))

                result[yy][xx] = (producer_storage[yy & 1][xx] +
                                  producer_storage[(yy + 1) & 1][xx] +
                                  producer_storage[yy & 1][xx + 1] +
                                  producer_storage[(yy + 1) & 1][xx + 1])

        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

        # The performance characteristics of this strategy are the
        # best so far. One of the four values of the producer we need
        # is probably still sitting in a register, so I won't count
        # it as a load:
        # producer.store_root().compute_at(consumer, x):
        # - Temporary memory allocated: 10 floats
        # - Loads: 48
        # - Stores: 56
        # - Calls to sqrt: 40

    # So what's the catch? Why not always do
    # producer.store_root().compute_at(consumer, x) for this type of
    # code?
    #
    # The answer is parallelism. In both of the previous two
    # strategies we've assumed that values computed on previous
    # iterations are lying around for us to reuse. This assumes that
    # previous values of x or y happened earlier in time and have
    # finished. This is not true if you parallelize or vectorize
    # either loop. Darn. If you parallelize, Halide won't inject the
    # optimizations that skip work already done if there's a parallel
    # loop in between the store_at level and the compute_at level,
    # and won't fold the storage down into a circular buffer either,
    # which makes our store_root pointless.

    # We're running out of options. We can make new ones by
    # splitting. We can store_at or compute_at at the natural
    # variables of the consumer (x and y), or we can split x or y
    # into new inner and outer sub-variables and then schedule with
    # respect to those. We'll use this to express fusion in tiles:
    if True:
        print("=" * 50)
        producer, consumer = hl.Func("producer_tile"), hl.Func("consumer_tile")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Tile the consumer using 2x2 tiles.
        x_outer, y_outer = hl.Var("x_outer"), hl.Var("y_outer")
        x_inner, y_inner = hl.Var("x_inner"), hl.Var("y_inner")
        consumer.tile(x, y, x_outer, y_outer, x_inner, y_inner, 2, 2)

        # Compute the producer per tile of the consumer
        producer.compute_at(consumer, x_outer)

        # Notice that I wrote my schedule starting from the end of
        # the pipeline (the consumer). This is because the schedule
        # for the producer refers to x_outer, which we introduced
        # when we tiled the consumer. You can write it in the other
        # order, but it tends to be harder to read.

        # Turn on tracing.
        producer.trace_stores()
        consumer.trace_stores()

        print("\nEvaluating:"
              "consumer.tile(x, y, x_outer, y_outer, x_inner, y_inner, 2, 2)"
              "producer.compute_at(consumer, x_outer)")
        consumer.realize([4, 4])

        # Reading the log, you should see that producer and consumer
        # now alternate on a per-tile basis. Here's the equivalent C:

        result = np.empty((4, 4), dtype=np.float32)

        # For every tile of the consumer:
        for y_outer in range(2):
            for x_outer in range(2):
                # Compute the x and y coords of the start of this tile.
                x_base = x_outer * 2
                y_base = y_outer * 2

                # Compute enough of producer to satisfy this tile. A
                # 2x2 tile of the consumer requires a 3x3 tile of the
                # producer.
                producer_storage = np.empty((3, 3), dtype=np.float32)
                for py in range(y_base, y_base + 3):
                    for px in range(x_base + 3):
                        producer_storage[py - y_base][px - x_base] = math.sqrt(px * py)

                # Compute this tile of the consumer
                for y_inner in range(2):
                    for x_inner in range(2):
                        xx = x_base + x_inner
                        yy = y_base + y_inner
                        result[yy][xx] = (producer_storage[yy - y_base][xx - x_base] +
                                          producer_storage[yy - y_base + 1][xx - x_base] +
                                          producer_storage[yy - y_base][xx - x_base + 1] +
                                          producer_storage[yy - y_base + 1][xx - x_base + 1])

        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

        # Tiling can make sense for problems like this one with
        # stencils that reach outwards in x and y. Each tile can be
        # computed independently in parallel, and the redundant work
        # done by each tile isn't so bad once the tiles get large
        # enough.

    # Let's try a mixed strategy that combines what we have done with
    # splitting, parallelizing, and vectorizing. This is one that
    # often works well in practice for large images. If you
    # understand this schedule, then you understand 95% of scheduling
    # in Halide.
    if True:
        print("=" * 50)
        producer, consumer = hl.Func("producer_mixed"), hl.Func("consumer_mixed")
        producer[x, y] = hl.sqrt(x * y)
        consumer[x, y] = (producer[x, y] +
                          producer[x, y + 1] +
                          producer[x + 1, y] +
                          producer[x + 1, y + 1])

        # Split the y coordinate of the consumer into strips of 16 scanlines:
        yo, yi = hl.Var("yo"), hl.Var("yi")
        consumer.split(y, yo, yi, 16)
        # Compute the strips using a thread pool and a task queue.
        consumer.parallel(yo)
        # Vectorize across x by a factor of four.
        consumer.vectorize(x, 4)

        # Now store the producer per-strip. This will be 17 scanlines
        # of the producer (16+1), but hopefully it will fold down
        # into a circular buffer of two scanlines:
        producer.store_at(consumer, yo)
        # Within each strip, compute the producer per scanline of the
        # consumer, skipping work done on previous scanlines.
        producer.compute_at(consumer, yi)
        # Also vectorize the producer (because sqrt is vectorizable on x86
        # using SSE).
        producer.vectorize(x, 4)

        # Let's leave tracing off this time, because we're going to
        # evaluate over a larger image.
        # consumer.trace_stores()
        # producer.trace_stores()

        halide_result = consumer.realize([800, 600])

        # Here's the equivalent (serial) C:

        c_result = np.empty((600, 800), dtype=np.float32)

        # For every strip of 16 scanlines
        for yo in range(600 // 16 + 1):  # (this loop is parallel in the Halide version)
            # 16 doesn't divide 600, so push the last slice upwards to fit
            # within [0, 599] (see lesson 05).
            y_base = yo * 16
            if y_base > (600 - 16):
                y_base = 600 - 16

            # Allocate a two-scanline circular buffer for the producer
            producer_storage = np.empty((2, 801), dtype=np.float32)

            # For every scanline in the strip of 16:
            for yi in range(16):
                yy = y_base + yi

                for py in range(yy, yy + 2):
                    # Skip scanlines already computed *within this task*
                    if (yi > 0) and (py == yy):
                        continue

                    # Compute this scanline of the producer in 4-wide vectors
                    for x_vec in range(800 // 4 + 1):
                        x_base = x_vec * 4
                        # 4 doesn't divide 801, so push the last vector left
                        # (see lesson 05).
                        if x_base > (801 - 4):
                            x_base = 801 - 4

                        # If you're on x86, Halide generates SSE code for this
                        # part:
                        xx = [x_base + 0, x_base + 1, x_base + 2, x_base + 3]
                        vec = [math.sqrt(xx[0] * py),
                               math.sqrt(xx[1] * py),
                               math.sqrt(xx[2] * py),
                               math.sqrt(xx[3] * py)]
                        producer_storage[py & 1][xx[0]] = vec[0]
                        producer_storage[py & 1][xx[1]] = vec[1]
                        producer_storage[py & 1][xx[2]] = vec[2]
                        producer_storage[py & 1][xx[3]] = vec[3]

                # Now compute consumer for this scanline:
                for x_vec in range(800 // 4):
                    x_base = x_vec * 4
                    # Again, Halide's equivalent here uses SSE.
                    xx = [x_base, x_base + 1, x_base + 2, x_base + 3]
                    vec = [
                        (producer_storage[yy & 1][xx[0]] +
                         producer_storage[(yy + 1) & 1][xx[0]] +
                         producer_storage[yy & 1][xx[0] + 1] +
                         producer_storage[(yy + 1) & 1][xx[0] + 1]),
                        (producer_storage[yy & 1][xx[1]] +
                         producer_storage[(yy + 1) & 1][xx[1]] +
                         producer_storage[yy & 1][xx[1] + 1] +
                         producer_storage[(yy + 1) & 1][xx[1] + 1]),
                        (producer_storage[yy & 1][xx[2]] +
                         producer_storage[(yy + 1) & 1][xx[2]] +
                         producer_storage[yy & 1][xx[2] + 1] +
                         producer_storage[(yy + 1) & 1][xx[2] + 1]),
                        (producer_storage[yy & 1][xx[3]] +
                         producer_storage[(yy + 1) & 1][xx[3]] +
                         producer_storage[yy & 1][xx[3] + 1] +
                         producer_storage[(yy + 1) & 1][xx[3] + 1])
                    ]

                    c_result[yy][xx[0]] = vec[0]
                    c_result[yy][xx[1]] = vec[1]
                    c_result[yy][xx[2]] = vec[2]
                    c_result[yy][xx[3]] = vec[3]

        print("Pseudo-code for the schedule:")
        consumer.print_loop_nest()
        print()

        # Look on my code, ye mighty, and despair!

        # Let's check the C result against the Halide result. Doing
        # this I found several bugs in my C implementation, which
        # should tell you something.
        for yy in range(600):
            for xx in range(800):
                error = halide_result[xx, yy] - c_result[yy][xx]
                # It's floating-point math, so we'll allow some slop:
                assert abs(error) <= 0.001, \
                    "halide_result(%d, %d) = %f instead of %f" % (
                        xx, yy, halide_result[xx, yy], c_result[yy][xx])

    # This stuff is hard. We ended up in a three-way trade-off
    # between memory bandwidth, redundant work, and
    # parallelism. Halide can't make the correct choice for you
    # automatically (sorry). Instead it tries to make it easier for
    # you to explore various options, without messing up your
    # program. In fact, Halide promises that scheduling calls like
    # compute_root won't change the meaning of your algorithm -- you
    # should get the same bits back no matter how you schedule
    # things.

    # So be empirical! Experiment with various schedules and keep a
    # log of performance. Form hypotheses and then try to prove
    # yourself wrong. Don't assume that you just need to vectorize
    # your code by a factor of four and run it on eight cores and
    # you'll get 32x faster. This almost never works. Modern systems
    # are complex enough that you can't predict performance reliably
    # without running your code.

    # We suggest you start by scheduling all of your non-trivial
    # stages compute_root, and then work from the end of the pipeline
    # upwards, inlining, parallelizing, and vectorizing each stage in
    # turn until you reach the top.

    # Halide is not just about vectorizing and parallelizing your
    # code. That's not enough to get you very far. Halide is about
    # giving you tools that help you quickly explore different
    # trade-offs between locality, redundant work, and parallelism,
    # without messing up the actual result you're trying to compute.

    print("=" * 50)
    print("Success!")
    return 0


if __name__ == "__main__":
    main()



import tensorflow as tf
import time

with tf.device('/GPU:0'):

    img = tf.random.uniform([4, 112, 112, 32])
    depthwise_filter = tf.random.uniform([3, 3, 32, 1])
    pointwise_filter = tf.random.uniform([1, 1, 32 * 1, 16])
    
    best = None
    num_trials = 10
    num_iter = 10
    for j in range(num_trials):
        start = time.time()
        for i in range(num_iter):
            out = tf.nn.separable_conv2d(
                img, depthwise_filter, pointwise_filter,
                strides = (1, 1, 1, 1), padding = 'VALID')
        end = time.time()
        t = (end - start) / num_iter
        if not best or t < best: best = t
    print('time: {} ms'.format(1000 * best))



"""
Local Laplacian.
"""

import halide as hl

# Just declare these at global scope, for simplicity
x, y, c, k = hl.vars("x y c k")


def _func_list(name, size):
    """Return a list containing `size` Funcs, named `name_n` for n in 0..size-1."""
    return [hl.Func("%s_%d" % (name, i)) for i in range(size)]


def _downsample(f):
    """Downsample with a 1 3 3 1 filter"""
    downx, downy = hl.funcs("downx downy")
    downx[x, y, hl._] = (
        f[2 * x - 1, y, hl._]
        + 3.0 * (f[2 * x, y, hl._] + f[2 * x + 1, y, hl._])
        + f[2 * x + 2, y, hl._]
    ) / 8.0
    downy[x, y, hl._] = (
        downx[x, 2 * y - 1, hl._]
        + 3.0 * (downx[x, 2 * y, hl._] + downx[x, 2 * y + 1, hl._])
        + downx[x, 2 * y + 2, hl._]
    ) / 8.0
    return downy


def _upsample(f):
    """Upsample using bilinear interpolation"""
    upx, upy = hl.funcs("upx upy")
    upx[x, y, hl._] = hl.lerp(
        f[(x + 1) // 2, y, hl._],
        f[(x - 1) // 2, y, hl._],
        ((x % 2) * 2 + 1) / 4.0,
    )
    upy[x, y, hl._] = hl.lerp(
        upx[x, (y + 1) // 2, hl._],
        upx[x, (y - 1) // 2, hl._],
        ((y % 2) * 2 + 1) / 4.0,
    )
    return upy


@hl.alias(local_laplacian_Mullapudi2016={"autoscheduler": "Mullapudi2016"})
@hl.generator()
class local_laplacian:
    pyramid_levels = hl.GeneratorParam(8)

    input_buf = hl.InputBuffer(hl.UInt(16), 3)
    levels = hl.InputScalar(hl.Int(32))
    alpha = hl.InputScalar(hl.Float(32))
    beta = hl.InputScalar(hl.Float(32))
    output_buf = hl.OutputBuffer(hl.UInt(16), 3)

    def generate(self):
        g = self

        # THE ALGORITHM
        J = g.pyramid_levels

        # Make the remapping function as a lookup table.
        fx = hl.f32(x) / 256.0
        remap = hl.Func("remap")
        remap[x] = g.alpha * fx * hl.exp(-fx * fx / 2.0)

        # Set a boundary condition
        clamped = hl.BoundaryConditions.repeat_edge(g.input_buf)

        # Convert to floating point
        floating = hl.Func("floating")
        floating[x, y, c] = clamped[x, y, c] / 65535.0

        # Get the luminance channel
        gray = hl.Func("gray")
        gray[x, y] = (
            hl.f32(0.299) * floating[x, y, 0]
            + hl.f32(0.587) * floating[x, y, 1]
            + hl.f32(0.114) * floating[x, y, 2]
        )

        # Make the processed Gaussian pyramid.
        gPyramid = _func_list("gPyramid", J)
        # Do a lookup into a lut with 256 entires per intensity level
        level = k * (1.0 / (g.levels - 1))
        idx = gray[x, y] * hl.f32(g.levels - 1) * 256.0
        idx = hl.clamp(hl.i32(idx), 0, (g.levels - 1) * 256)
        gPyramid[0][x, y, k] = (
            g.beta * (gray[x, y] - level) + level + remap[idx - 256 * k]
        )
        for j in range(1, J):
            gPyramid[j][x, y, k] = _downsample(gPyramid[j - 1])[x, y, k]

        # Get its laplacian pyramid
        lPyramid = _func_list("lPyramid", J)
        lPyramid[J - 1][x, y, k] = gPyramid[J - 1][x, y, k]
        for j in range(J - 2, -1, -1):
            lPyramid[j][x, y, k] = (
                gPyramid[j][x, y, k] - _upsample(gPyramid[j + 1])[x, y, k]
            )

        # Make the Gaussian pyramid of the input
        inGPyramid = _func_list("inGPyramid", J)
        inGPyramid[0][x, y] = gray[x, y]
        for j in range(1, J):
            inGPyramid[j][x, y] = _downsample(inGPyramid[j - 1])[x, y]

        # Make the laplacian pyramid of the output
        outLPyramid = _func_list("outLPyramid", J)
        for j in range(0, J):
            # Split input pyramid value into integer and floating parts
            level = inGPyramid[j][x, y] * hl.f32(g.levels - 1)
            li = hl.clamp(hl.i32(level), 0, g.levels - 2)
            lf = level - hl.f32(li)
            # Linearly interpolate between the nearest processed pyramid levels
            outLPyramid[j][x, y] = (1.0 - lf) * lPyramid[j][x, y, li] + (
                lf * lPyramid[j][x, y, li + 1]
            )

        # Make the Gaussian pyramid of the output
        outGPyramid = _func_list("outGPyramid", J)
        outGPyramid[J - 1][x, y] = outLPyramid[J - 1][x, y]
        for j in range(J - 2, -1, -1):
            outGPyramid[j][x, y] = (
                _upsample(outGPyramid[j + 1])[x, y] + outLPyramid[j][x, y]
            )

        # Reintroduce color (Connelly: use eps to avoid scaling up noise w/
        # apollo3.png input)
        color = hl.Func("color")
        eps = hl.f32(0.01)
        color[x, y, c] = (
            outGPyramid[0][x, y] * (floating[x, y, c] + eps) / (gray[x, y] + eps)
        )

        # Convert back to 16-bit
        g.output_buf[x, y, c] = hl.u16(hl.clamp(color[x, y, c], 0.0, 1.0) * 65535.0)

        # ESTIMATES
        # (This can be useful in conjunction with RunGen and benchmarks as well
        # as autoschedulers, so we do it in all cases.)
        g.input_buf.set_estimates([(0, 1536), (0, 2560), (0, 3)])
        # Provide estimates on the parameters
        g.levels.set_estimate(8)
        g.alpha.set_estimate(1)
        g.beta.set_estimate(1)
        g.output_buf.set_estimates([(0, 1536), (0, 2560), (0, 3)])

        # THE SCHEDULE
        if g.using_autoscheduler():
            # nothing
            pass
        elif g.target().has_gpu_feature():
            # GPU schedule.
            # 3.19ms on an RTX 2060.
            remap.compute_root()
            xi, yi = hl.vars("xi yi")
            g.output_buf.compute_root().gpu_tile(x, y, xi, yi, 16, 8)
            for j in range(0, J):
                blockw = 16
                blockh = 8
                if j > 3:
                    blockw = 2
                    blockh = 2

                if j > 0:
                    inGPyramid[j].compute_root().gpu_tile(x, y, xi, yi, blockw, blockh)
                    (
                        gPyramid[j]
                        .compute_root()
                        .reorder(k, x, y)
                        .gpu_tile(x, y, xi, yi, blockw, blockh)
                    )

                outGPyramid[j].compute_root().gpu_tile(x, y, xi, yi, blockw, blockh)

        else:
            # CPU schedule.

            # 21.4ms on an Intel i9-9960X using 32 threads at 3.7
            # GHz, using the target x86-64-avx2.

            # This app is dominated by data-dependent loads from
            # memory, so we're better off leaving the AVX-512 units
            # off in exchange for a higher clock, and we benefit from
            # hyperthreading.

            remap.compute_root()
            yo = hl.Var("yo")
            (
                g.output_buf.reorder(c, x, y)
                .split(y, yo, y, 64)
                .parallel(yo)
                .vectorize(x, 8)
            )
            gray.compute_root().parallel(y, 32).vectorize(x, 8)
            for j in range(1, 5):
                inGPyramid[j].compute_root().parallel(y, 32).vectorize(x, 8)
                (
                    gPyramid[j]
                    .compute_root()
                    .reorder_storage(x, k, y)
                    .reorder(k, y)
                    .parallel(y, 8)
                    .vectorize(x, 8)
                )
                (
                    outGPyramid[j]
                    .store_at(g.output_buf, yo)
                    .compute_at(g.output_buf, y)
                    .fold_storage(y, 4)
                    .vectorize(x, 8)
                )

            outGPyramid[0].compute_at(g.output_buf, y).vectorize(x, 8)
            for j in range(5, J):
                inGPyramid[j].compute_root()
                gPyramid[j].compute_root().parallel(k)
                outGPyramid[j].compute_root()


if __name__ == "__main__":
    hl.main()



from .halide_ import *
from .halide_ import _, _0, _1, _2, _3, _4, _5, _6, _7, _8, _9
from ._generator_helpers import (
    _create_python_generator,
    _generatorcontext_enter,
    _generatorcontext_exit,
    _get_python_generator_names,
    active_generator_context,
    alias,
    funcs,
    Generator,
    generator,
    GeneratorParam,
    InputBuffer,
    InputScalar,
    OutputBuffer,
    OutputScalar,
    vars,
)



"""
Bilateral histogram.
"""

from bilateral_grid import bilateral_grid
from bilateral_grid_Adams2019 import bilateral_grid_Adams2019
from bilateral_grid_Li2018 import bilateral_grid_Li2018
from bilateral_grid_Mullapudi2016 import bilateral_grid_Mullapudi2016
import halide.imageio
import numpy as np
import sys
import timeit


def main():
    if len(sys.argv) < 4:
        print("Usage: %s input.png output.png range_sigma" % sys.argv[0])
        print("e.g. %s input.png output.png 0.1 10" % sys.argv[0])
        sys.exit(1)

    input_path = sys.argv[1]
    r_sigma = float(sys.argv[2])
    output_path = sys.argv[3]
    timing_iterations = 10

    print("Reading from %s ..." % input_path)
    input_buf_u8 = halide.imageio.imread(input_path)
    assert input_buf_u8.dtype == np.uint8
    # Convert to float32
    input_buf = input_buf_u8.astype(np.float32)
    input_buf /= 255.0
    output_buf = np.empty(input_buf.shape, dtype=input_buf.dtype)

    tests = {
        "Manual": bilateral_grid,
        "Adams2019": bilateral_grid_Adams2019,
        #
        # TODO: Don't test the Li2018-autoscheduled version here;
        # it schedules the histogram stage with compute_root(),
        # which (for a 4MP input image) attempts to allocate 124GB (!)...
        #
        # "Li2018": bilateral_grid_Li2018,
        #
        "Mullapudi2016": bilateral_grid_Mullapudi2016,
    }

    for name, fn in tests.items():
        print("Running %s... " % name, end="")
        t = timeit.Timer(lambda: fn(input_buf, r_sigma, output_buf))
        avg_time_sec = t.timeit(number=timing_iterations) / timing_iterations
        print("time: %fms" % (avg_time_sec * 1e3))

    output_buf *= 255.0
    output_buf_u8 = output_buf.astype(np.uint8)
    print("Saving to %s ..." % output_path)
    halide.imageio.imwrite(output_path, output_buf_u8)

    print("Success!")
    sys.exit(0)


if __name__ == "__main__":
    main()



import torch
import torchvision.models.resnet as resnet
import torchvision.transforms as transforms
import torch.nn as nn
import numpy as np
import sys

if __name__ == "__main__":
    assert(len(sys.argv) == 3)
    halide_output_file = sys.argv[1]

    seed = int(sys.argv[2])
    np.random.seed(seed)
    input_data = np.random.rand(1, 224, 224, 3).astype(np.float32)
    input_data = input_data.transpose(0,3,1,2)
    image = torch.from_numpy(input_data)

    net = resnet.resnet50(pretrained=True) 
    net.eval()
    result = net(image)
    # PyTorch resnet50 model doesn't compute softmax of the output.
    softmax = nn.Softmax(dim=1)
    ref_output = softmax(result)

    halide_output = np.fromfile(halide_output_file, dtype=np.float32)
    
    ref_output = ref_output.cpu().detach().numpy().reshape(halide_output.shape)

    np.testing.assert_almost_equal(np.argmax(halide_output), np.argmax(ref_output))
    np.testing.assert_almost_equal(halide_output, ref_output, decimal=2)
    print('Success!')



import halide as hl
import numpy as np


def test_extern():
    """
    Shows an example of Halide calling a C library loaded
    in the Python process via ctypes
    """

    # Requires Makefile support to build the external function in linkable form
    print("TODO: test_extern not yet implemented in Python; skipping...")
    return 0

    x = hl.Var("x")

    data = np.random.random(10).astype(np.float64)
    expected_result = np.sort(data)
    output_data = np.empty(10, dtype=np.float64)

    sort_func = hl.Func("extern_sort_func")
    # gsl_sort,
    # see http://www.gnu.org/software/gsl/manual/html_node/Sorting-vectors.html#Sorting-vectors

    input = hl.ImageParam(hl.Float(64), 1, "input_data")

    extern_name = "the_sort_func"
    params = [hl.ExternFuncArgument(input)]
    output_types = [hl.Int(32)]
    dimensionality = 1
    sort_func.define_extern(extern_name, params, output_types, dimensionality)

    try:
        sort_func.compile_jit()
    except hl.HalideError:
        assert "cannot be converted to a bool" in str(e)
    else:
        assert False, "Did not see expected exception!"

    import ctypes

    sort_lib = ctypes.CDLL("the_sort_function.so")
    print(sort_lib.the_sort_func)

    try:
        sort_func.compile_jit()
    except hl.HalideError:
        assert "cannot be converted to a bool" in str(e)
    else:
        assert False, "Did not see expected exception!"

    lib_path = "the_sort_function.so"
    load_error = load_library_into_llvm(lib_path)
    assert load_error == False

    sort_func.compile_jit()

    # now that things are loaded, we try to call them
    input.set(data)
    sort_func.realize(output_data)

    assert np.isclose(expected_result, output_data)

    return


if __name__ == "__main__":

    test_extern()



import halide as hl


def test_var():
    v1 = hl.Var()
    v2 = hl.Var()
    assert len(v1.name()) > 0
    assert len(v2.name()) > 0
    assert not v1.same_as(v2)

    v1 = hl.Var.implicit(1)
    assert v1.name() == "_1"
    v2 = hl.Var("_1")
    assert v1.same_as(v2)
    v3 = hl._1
    assert v1.same_as(v3)
    v4 = hl.Var("v4")
    assert not v1.same_as(v4)

    assert v1.is_implicit()
    assert v2.is_implicit()
    assert v3.is_implicit()
    assert not v4.is_implicit()

    assert hl.Var("_1").is_implicit()
    assert not hl.Var("v4").is_implicit()

    assert v1.implicit_index() == 1
    assert v2.implicit_index() == 1
    assert v3.implicit_index() == 1
    assert v4.implicit_index() == -1

    assert hl.Var("_1").implicit_index() == 1
    assert hl.Var("v4").implicit_index() == -1

    ph = hl._
    assert ph.name() == "_"
    assert ph.is_placeholder()
    assert hl.Var.is_placeholder(ph)
    assert not v1.is_placeholder()

    outermost = hl.Var.outermost()
    assert outermost.name() == "__outermost"

    # repr() and str()
    x = hl.Var("x")
    assert str(x) == "x"
    assert repr(x) == "<halide.Var 'x'>"

    # This verifies that halide.Var is implicitly convertible to halide.Expr
    r = hl.random_int(x)

    # This verifies that halide.Var is explicitly convertible to halide.Expr
    r = hl.random_int(hl.Expr(x))


if __name__ == "__main__":
    test_var()



import halide as hl


def test_autodiff():
    x = hl.Var("x")
    b = hl.Buffer(hl.Float(32), [3])
    p = hl.Param(hl.Float(32), "p", 1)
    b[0] = 1.0
    b[1] = 2.0
    b[2] = 3.0
    f, g, h = hl.Func("f"), hl.Func("g"), hl.Func("h")
    f[x] = b[x]
    f[0] = 4.0
    g[x] = f[x] * 5.0 * p
    r = hl.RDom([(0, 3)])
    h[()] = 0.0
    h[()] += g[r.x]

    d = hl.propagate_adjoints(h)

    # gradient w.r.t. the initialization of f
    d_f_init = d[f]
    d_f_init_buf = d_f_init.realize([3])
    assert d_f_init_buf[0] == 0.0
    assert d_f_init_buf[1] == 5.0
    assert d_f_init_buf[2] == 5.0
    d_f_init = d[f]  # test different interface
    d_f_init_buf = d_f_init.realize([3])
    assert d_f_init_buf[0] == 0.0
    assert d_f_init_buf[1] == 5.0
    assert d_f_init_buf[2] == 5.0

    # gradient w.r.t. the updated f
    d_f_update_0 = d[f, 0]
    d_f_update_0_buf = d_f_update_0.realize([3])
    assert d_f_update_0_buf[0] == 5.0
    assert d_f_update_0_buf[1] == 5.0
    assert d_f_update_0_buf[2] == 5.0
    d_f_update_0 = d[f, 0]
    d_f_update_0_buf = d_f_update_0.realize([3])
    assert d_f_update_0_buf[0] == 5.0
    assert d_f_update_0_buf[1] == 5.0
    assert d_f_update_0_buf[2] == 5.0

    # gradient w.r.t. the buffer
    d_b = d[b]
    d_b_buf = d_b.realize([3])
    assert d_b_buf[0] == 0.0
    assert d_b_buf[1] == 5.0
    assert d_b_buf[2] == 5.0
    d_b = d[b]
    d_b_buf = d_b.realize([3])
    assert d_b_buf[0] == 0.0
    assert d_b_buf[1] == 5.0
    assert d_b_buf[2] == 5.0

    # gradient w.r.t. the param
    d_p = d[p]
    d_p_buf = d_p.realize()
    # 5 * (4 + 2 + 3)
    assert abs(d_p_buf[()] - 45.0) < 1e-6
    d_p = d[p]
    d_p_buf = d_p.realize()
    assert abs(d_p_buf[()] - 45.0) < 1e-6


if __name__ == "__main__":
    test_autodiff()



"""Verifies the Halide operator functions properly."""


import os
import unittest
import warnings

import torch as th
import modules


class TestAdd(unittest.TestCase):
    def setUp(self):
        self.a = th.ones(1, 2, 8, 8)
        self.b = th.ones(1, 2, 8, 8)*3
        self.gt = th.ones(1, 2, 8, 8)*4

    def test_cpu_single(self):
        self._test_add(is_double=False)

    def test_cpu_double(self):
        self._test_add(is_double=True)

    def test_gpu_single(self):
        if not th.cuda.is_available():
            return
        self._test_add(is_cuda=True, is_double=False)

    def test_gpu_double(self):
        if not th.cuda.is_available():
            return
        self._test_add(is_cuda=True, is_double=True)

    def _test_add(self, is_cuda=False, is_double=False):
        if is_double:
            self.a = self.a.double()
            self.b = self.b.double()
            self.gt = self.gt.double()
        if is_cuda:
            print("Testing Halide PyTorch CUDA operator...")
            self.a = self.a.cuda()
            self.b = self.b.cuda()
            self.gt = self.gt.cuda()
        else:
            print("Testing Halide PyTorch CPU operator...")

        for backward_op in ["add_grad", "add_halidegrad"]:
            add = modules.Add(backward_op)
            output = add(self.a, self.b)

            if is_double:
                print("  .Double-precision mode, backward_op:", backward_op)
            else:
                print("  .Single-precision mode, backward_op:", backward_op)

            diff = (output-self.gt).sum().item()
            assert diff == 0.0, "Test failed: sum should be 4, got %f" % diff

            self.a.requires_grad = True
            self.b.requires_grad = True

            for i in range(100):
                output = add(self.a, self.b).sum()
                output.backward()
            
            # Inputs are float, the gradient checker wants double inputs and
            # will issue a warning.
            warnings.filterwarnings(
                "ignore", module=r".*gradcheck*")

            # Test the gradient is correct
            res = th.autograd.gradcheck(add, [self.a, self.b], eps=1e-2)

            print("     Test ran successfully: difference is", diff)


if __name__ == "__main__":
    unittest.main()



import halide as hl


def test_free_logical_not_function():
    x = hl.Var("x")

    f = hl.Func("f")
    f[x] = x > 5

    not_f = hl.Func("not_f")
    not_f[x] = hl.logical_not(f[x])

    f_out = f.realize([10])
    not_f_out = not_f.realize([10])

    for i in range(10):
        assert f_out[i] == (i > 5)
        assert not_f_out[i] == (i <= 5)


def test_member_logical_not_function():
    x = hl.Var("x")

    f = hl.Func("f")
    f[x] = x > 5

    not_f = hl.Func("not_f")
    not_f[x] = f[x].logical_not()

    f_out = f.realize([10])
    not_f_out = not_f.realize([10])

    for i in range(10):
        assert f_out[i] == (i > 5)
        assert not_f_out[i] == (i <= 5)


if __name__ == "__main__":
    test_free_logical_not_function()
    test_member_logical_not_function()



import halide as hl


def test_multipass_constraints():
    input = hl.ImageParam(hl.Float(32), 2, "input")

    f = hl.Func("f")
    x = hl.Var("x")
    y = hl.Var("y")

    f[x, y] = input[x + 1, y + 1] + input[x - 1, y - 1]
    f[x, y] += 3.0
    f.update().vectorize(x, 4)

    o = f.output_buffer()

    # Now make some hard-to-resolve constraints
    input.dim(0).set_bounds(
        min=input.dim(1).min() - 5, extent=input.dim(1).extent() + o.dim(0).extent()
    )

    o.dim(0).set_bounds(
        min=0,
        extent=hl.select(
            o.dim(0).extent() < 22, o.dim(0).extent() + 1, o.dim(0).extent()
        ),
    )

    # Make a bounds query buffer
    query_buf = hl.Buffer.make_bounds_query(type=hl.Float(32), sizes=[7, 8])
    query_buf.set_min([2, 2])

    f.infer_input_bounds(query_buf)

    if (
        input.get().dim(0).min() != -4
        or input.get().dim(0).extent() != 34
        or input.get().dim(1).min() != 1
        or input.get().dim(1).extent() != 10
        or query_buf.dim(0).min() != 0
        or query_buf.dim(0).extent() != 24
        or query_buf.dim(1).min() != 2
        or query_buf.dim(1).extent() != 8
    ):

        print(
            "Constraints not correctly satisfied:\n",
            "in:",
            input.get().dim(0).min(),
            input.get().dim(0).extent(),
            input.get().dim(1).min(),
            input.get().dim(1).extent(),
            "out:",
            query_buf.dim(0).min(),
            query_buf.dim(0).extent(),
            query_buf.dim(1).min(),
            query_buf.dim(1).extent(),
        )
        assert False


if __name__ == "__main__":
    test_multipass_constraints()



import halide as hl

import simplecpp_pystub
import complexcpp_pystub

from simplepy_generator import SimplePy
from complexpy_generator import ComplexPy


def _realize_and_check(f, offset=0):
    b = hl.Buffer(hl.Float(32), [2, 2])
    f.realize(b)

    assert b[0, 0] == 3.5 + offset + 123
    assert b[0, 1] == 4.5 + offset + 123
    assert b[1, 0] == 4.5 + offset + 123
    assert b[1, 1] == 5.5 + offset + 123


def test_simple(cls):
    x, y = hl.Var(), hl.Var()

    b_in = hl.Buffer(hl.UInt(8), [2, 2])
    b_in.fill(123)
    for xx in range(2):
        for yy in range(2):
            b_in[xx, yy] += xx + yy

    # ----------- Inputs by-position
    f = cls.call(b_in, 3.5)
    _realize_and_check(f)

    # ----------- Inputs by-name
    f = cls.call(buffer_input=b_in, float_arg=3.5)
    _realize_and_check(f)

    f = cls.call(float_arg=3.5, buffer_input=b_in)
    _realize_and_check(f)

    # ----------- Above set again, w/ GeneratorParam mixed in
    k = 42

    gp = {"offset": k}

    # (positional)
    f = cls.call(b_in, 3.5, generator_params=gp)
    _realize_and_check(f, k)

    # (keyword)
    f = cls.call(generator_params=gp, buffer_input=b_in, float_arg=3.5)
    _realize_and_check(f, k)

    f = cls.call(buffer_input=b_in, generator_params=gp, float_arg=3.5)
    _realize_and_check(f, k)

    f = cls.call(buffer_input=b_in, generator_params=gp, float_arg=3.5)
    _realize_and_check(f, k)

    f = cls.call(buffer_input=b_in, float_arg=3.5, generator_params=gp)

    # Inputs w/ mixed by-position and by-name should be ok
    f = cls.call(b_in, float_arg=3.5, generator_params=gp)
    _realize_and_check(f, k)

    # ----------- Test various failure modes
    try:
        # too many positional args
        f = cls.call(b_in, 3.5, 4)
    except hl.HalideError as e:
        assert "allows at most 2 positional args, but 3 were specified." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # too few positional args
        f = cls.call(b_in)
    except hl.HalideError as e:
        assert "requires 2 args, but 1 were specified." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Inputs that can't be converted to what the receiver needs (positional)
        f = cls.call(hl.f32(3.141592), "happy")
    except hl.HalideError as e:
        assert "Input buffer_input requires an ImageParam or Buffer argument" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Inputs that can't be converted to what the receiver needs (named)
        f = cls.call(b_in, float_arg="bogus")
    except hl.HalideError as e:
        assert (
            "Input float_arg requires a Param (or scalar literal) argument when using call"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"

    try:
        # Input specified by both pos and kwarg
        f = cls.call(b_in, 3.5, float_arg=4.5)
    except hl.HalideError as e:
        assert "Input float_arg specified multiple times." in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # generator_params is not a dict
        f = cls.call(b_in, 3.5, generator_params=[1, 2, 3])
    except hl.HalideError as e:
        assert "generator_params must be a dict" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Bad gp name
        f = cls.call(b_in, 3.5, generator_params={"foo": 0})
    except hl.HalideError as e:
        assert "has no GeneratorParam" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Bad input name
        f = cls.call(buzzer_input=b_in, float_arg=3.5, generator_params=gp)
    except hl.HalideError as e:
        assert "Unknown input 'buzzer_input' specified via keyword argument" in str(e)
    else:
        assert False, "Did not see expected exception!"

    try:
        # Bad gp name
        f = cls.call(
            buffer_input=b_in,
            float_arg=3.5,
            generator_params=gp,
            nonexistent_generator_param="wat",
        )
    except hl.HalideError as e:
        assert (
            "Unknown input 'nonexistent_generator_param' specified via keyword argument"
            in str(e)
        )
    else:
        assert False, "Did not see expected exception!"


def _make_constant_image(type):
    constant_image = hl.Buffer(type, [32, 32, 3], "constant_image")
    for x in range(32):
        for y in range(32):
            for c in range(3):
                constant_image[x, y, c] = x + y + c
    return constant_image


def test_complex(cls, extra_input_name=""):
    constant_image = _make_constant_image(hl.UInt(8))
    constant_image_u16 = _make_constant_image(hl.UInt(16))
    input = hl.ImageParam(hl.UInt(8), 3, "input")
    input.set(constant_image)

    x, y, c = hl.Var(), hl.Var(), hl.Var()

    float_arg = 1.25
    int_arg = 33
    gp = {
        "simple_input.type": hl.UInt(8),
        "tuple_output.type": [hl.Float(32), hl.Float(32)],
        "untyped_buffer_input.type": hl.UInt(8),
        "untyped_buffer_output.dim": 3,
        "untyped_buffer_output.type": hl.UInt(8),
        "vectorize": True,
    }
    kwargs = {
        "typed_buffer_input": constant_image,
        "untyped_buffer_input": constant_image,
        "simple_input": constant_image,
        "float_arg": float_arg,
        "int_arg": int_arg,
        "generator_params": gp,
    }
    if len(extra_input_name):
        gp["extra_input_name"] = extra_input_name
        kwargs[extra_input_name] = constant_image_u16

    r = cls.call(**kwargs)

    # return value is a tuple; unpack separately to avoid
    # making the callsite above unreadable
    (
        simple_output,
        tuple_output,
        typed_buffer_output,
        untyped_buffer_output,
        static_compiled_buffer_output,
        scalar_output,
        extra_func_output,
    ) = r

    b = simple_output.realize([32, 32, 3], target)
    assert b.type() == hl.Float(32)
    for x in range(32):
        for y in range(32):
            for c in range(3):
                expected = constant_image[x, y, c]
                actual = b[x, y, c]
                assert expected == actual, "Expected %s Actual %s" % (expected, actual)

    b = tuple_output.realize([32, 32, 3], target)
    assert b[0].type() == hl.Float(32)
    assert b[1].type() == hl.Float(32)
    assert len(b) == 2
    for x in range(32):
        for y in range(32):
            for c in range(3):
                expected1 = constant_image[x, y, c] * float_arg
                expected2 = expected1 + int_arg
                actual1, actual2 = b[0][x, y, c], b[1][x, y, c]
                assert expected1 == actual1, "Expected1 %s Actual1 %s" % (
                    expected1,
                    actual1,
                )
                assert expected2 == actual2, "Expected2 %s Actual1 %s" % (
                    expected2,
                    actual2,
                )

    # TODO: Output<Buffer<>> has additional behaviors useful when a Stub
    # is used within another Generator; this isn't yet implemented since there
    # isn't yet Python bindings for Generator authoring. This section
    # of the test may need revision at that point.
    b = typed_buffer_output.realize([32, 32, 3], target)
    assert b.type() == hl.Float(32)
    for x in range(32):
        for y in range(32):
            for c in range(3):
                expected = constant_image[x, y, c]
                actual = b[x, y, c]
                assert expected == actual, "Expected %s Actual %s" % (expected, actual)

    b = untyped_buffer_output.realize([32, 32, 3], target)
    assert b.type() == hl.UInt(8)
    for x in range(32):
        for y in range(32):
            for c in range(3):
                expected = constant_image[x, y, c]
                actual = b[x, y, c]
                assert expected == actual, "Expected %s Actual %s" % (expected, actual)

    b = static_compiled_buffer_output.realize([4, 4, 1], target)
    assert b.type() == hl.UInt(8)
    for x in range(4):
        for y in range(4):
            for c in range(1):
                expected = constant_image[x, y, c] + 42
                actual = b[x, y, c]
                assert expected == actual, "Expected %s Actual %s" % (expected, actual)

    b = scalar_output.realize([], target)
    assert b.type() == hl.Float(32)
    assert b[()] == 34.25

    b = extra_func_output.realize([32, 32], target)
    assert b.type() == hl.Float(64)
    for x in range(32):
        for y in range(32):
            if len(extra_input_name):
                expected = x + y + 1
            else:
                expected = 0
            actual = b[x, y]
            assert expected == actual, "Expected %s Actual %s" % (expected, actual)


if __name__ == "__main__":
    target = hl.get_jit_target_from_environment()
    with hl.GeneratorContext(target):
        test_simple(simplecpp_pystub)
        test_complex(complexcpp_pystub)
        test_complex(complexcpp_pystub, extra_input_name="foo_input")
        test_simple(SimplePy)
        test_complex(ComplexPy)
        test_complex(ComplexPy, extra_input_name="foo_input")



"""Backend for running ONNX using Halide"""


from onnx.backend.base import Backend as BackendBase
import onnx
import model as halide_model
import signal
import base64
import hashlib
import datetime


class HalideBackend(BackendBase):
    @classmethod
    def is_compatible(cls,
                      model,  # type: ModelProto
                      device='CPU',  # type: Text
                      **kwargs  # type: Any
                      ):  # type: (...) -> bool
        """Returns whether the model is compatible with the backend.
        For the moment, we always return True and will throw an exception
        later when preparing or running the model.
        """
        return True

    @classmethod
    def prepare(cls,
                model,  # type: ModelProto
                device='CPU',  # type: Text
                **kwargs  # type: Any
                ):
        """Prepare an ONNX model to run using the Halide backend.

        Builds and returns an internal representation of the model (which
        includes the actual Halide pipeline).

        :param model: The ONNX model to be converted.
        :param device: The device to execute this model on (Ignored for now).
        :returns: An internal object that can be used to run the model with
                  Halide.
        """
        onnx.checker.check_model(model)

        prepared = halide_model.Model()
        prepared.BuildFromOnnxModel(model)
        # Optimize the schedule of nontrivial models to make sure they
        # complete in a reasonable amount of time
        if len(model.graph.node) > 10:
            prepared.OptimizeSchedule()
        return prepared

    @classmethod
    def run_model(cls,
                  model,  # type: ModelProto
                  inputs,  # type: Any
                  device='CPU',  # type: Text
                  **kwargs  # type: Any
                  ):  # type: (...) -> Tuple[Any, ...]
        """Evaluate a ONNX model using the Halide backend.

        :param model: The ONNX model to be converted.
        :param inputs: A list of numpy arrays (one for each model input)
        :param device: The device to execute this model on (Ignored for now).
        :returns: A list of numpy arrays (one for each model output).
        """
        prepared = cls.prepare(model, device, **kwargs)
        return prepared.run(inputs, device)

    @classmethod
    def run_node(cls,
                 node,  # type: NodeProto
                 inputs,  # type: Any
                 device='CPU',  # type: Text
                 outputs_info=None,   # type: Optional[Sequence[Tuple[numpy.dtype, Tuple[int, ...]]]]
                 **kwargs  # type: Dict[Text, Any]
                 ):  # type: (...) -> Optional[Tuple[Any, ...]]
        """Evaluate a single ONNX node using the Halide backend.
        Not supported yet.
        """
        assert False
        return None

    @classmethod
    def supports_device(cls, device):  # type: (Text) -> bool
        """
        Checks whether the backend is compiled with support for the requested
        device and that device is available on the machine.
        """
        if device == 'CPU':
            return True
        else:
            # We don't support anything else at this time
            return False


prepare = HalideBackend.prepare
run_node = HalideBackend.run_node
run_model = HalideBackend.run_model
supports_device = HalideBackend.supports_device
is_compatible = HalideBackend.is_compatible




#!/usr/bin/python3

# Halide tutorial lesson 9

# This lesson demonstrates how to define a hl.Func in multiple passes,
# including scattering.

# This lesson can be built by invoking the command:
#    make test_tutorial_lesson_09_update_definitions
# in a shell with the current directory at python_bindings/
from datetime import datetime

import halide as hl

import halide.imageio
import numpy as np
import os.path


def main():
    # Declare some Vars to use below.
    x, y = hl.Var("x"), hl.Var("y")

    # Load a grayscale image to use as an input.
    image_path = os.path.join(os.path.dirname(__file__), "../../tutorial/images/gray.png")
    input_data = halide.imageio.imread(image_path)
    if True:
         # making the image smaller to go faster
        input_data = input_data[:150, :160]
    assert input_data.dtype == np.uint8
    input = hl.Buffer(input_data)

    # You can define a hl.Func in multiple passes. Let's see a toy
    # example first.
    if True:
        # The first definition must be one like we have seen already
        # - a mapping from Vars to an hl.Expr:
        f = hl.Func("f")
        f[x, y] = x + y
        # We call this first definition the "pure" definition.

        # But the later definitions can include computed expressions on
        # both sides. The simplest example is modifying a single point:
        f[3, 7] = 42

        # We call these extra definitions "update" definitions, or
        # "reduction" definitions. A reduction definition is an
        # update definition that recursively refers back to the
        # function's current value at the same site:
        if False:
            e = f[x, y] + 17
            print("f[x, y] + 17", e)
            print("(f[x, y] + 17).type()", e.type())
            print("(f[x, y]).type()", f[x, y].type())

        f[x, y] = f[x, y] + 17

        # If we confine our update to a single row, we can
        # recursively refer to values in the same column:
        f[x, 3] = f[x, 0] * f[x, 10]

        # Similarly, if we confine our update to a single column, we
        # can recursively refer to other values in the same row.
        f[0, y] = f[0, y] / f[3, y]

        # The general rule is: Each hl.Var used in an update definition
        # must appear unadorned in the same position as in the pure
        # definition in all references to the function on the left-
        # and right-hand sides. So the following definitions are
        # legal updates:
        # x is used, so all uses of f must have x as the first argument.
        f[x, 17] = x + 8
        # y is used, so all uses of f must have y as the second argument.
        f[0, y] = y * 8
        f[x, x + 1] = x + 8
        f[y / 2, y] = f[0, y] * 17

        # But these ones would cause an error:
        # f[x, 0) = f[x + 1, 0) <- First argument to f on the right-hand-side must be 'x', not 'x + 1'.
        # f[y, y + 1) = y + 8   <- Second argument to f on the left-hand-side must be 'y', not 'y + 1'.
        # f[y, x) = y - x      <- Arguments to f on the left-hand-side are in the wrong places.
        # f[3, 4) = x + y      <- Free variables appear on the right-hand-side
        # but not the left-hand-side.

        # We'll realize this one just to make sure it compiles. The
        # second-to-last definition forces us to realize over a
        # domain that is taller than it is wide.
        f.realize([100, 101])

        # For each realization of f, each step runs in its entirety
        # before the next one begins. Let's trace the loads and
        # stores for a simpler example:
        g = hl.Func("g")
        g[x, y] = x + y   # Pure definition
        g[2, 1] = 42      # First update definition
        g[x, 0] = g[x, 1]  # Second update definition

        g.trace_loads()
        g.trace_stores()

        g.realize([4, 4])

        # Reading the log, we see that each pass is applied in turn. The
        # equivalent Python is:
        result = np.empty((4, 4), dtype=np.int32)
        # Pure definition
        for yy in range(4):
            for xx in range(4):
                result[yy][xx] = xx + yy

        # First update definition
        result[1][2] = 42
        # Second update definition
        for xx in range(4):
            result[0][xx] = result[1][xx]
    # end of section

    # Putting update passes inside loops.
    if True:
        # Starting with this pure definition:
        f = hl.Func("f")
        f[x, y] = x + y

        # Say we want an update that squares the first fifty rows. We
        # could do this by adding 50 update definitions:

        # f[x, 0) = f[x, 0) * f[x, 0)
        # f[x, 1) = f[x, 1) * f[x, 1)
        # f[x, 2) = f[x, 2) * f[x, 2)
        # ...
        # f[x, 49) = f[x, 49) * f[x, 49)

        # Or equivalently using a compile-time loop in our C++:
        # for (int i = 0 i < 50 i++) {
        #   f[x, i) = f[x, i) * f[x, i)
        #

        # But it's more manageable and more flexible to put the loop
        # in the generated code. We do this by defining a "reduction
        # domain" and using it inside an update definition:
        r = hl.RDom([(0, 50)])
        f[x, r] = f[x, r] * f[x, r]
        halide_result = f.realize([100, 100])

        # The equivalent Python is:
        py_result = np.empty((100, 100), dtype=np.int32)
        for yy in range(100):
            for xx in range(100):
                py_result[yy][xx] = xx + yy

        for xx in range(100):
            for rr in range(50):
                # The loop over the reduction domain occurs inside of
                # the loop over any pure variables used in the update
                # step:
                py_result[rr][xx] = py_result[rr][xx] * py_result[rr][xx]

        # Check the results match:
        for yy in range(100):
            for xx in range(100):
                assert halide_result[xx, yy] == py_result[yy][xx], \
                    "halide_result(%d, %d) = %d instead of %d" % (
                        xx, yy, halide_result[xx, yy], py_result[yy][xx])

    # Now we'll examine a real-world use for an update definition:
    # computing a histogram.
    if True:

        # Some operations on images can't be cleanly expressed as a pure
        # function from the output coordinates to the value stored
        # there. The classic example is computing a histogram. The
        # natural way to do it is to iterate over the input image,
        # updating histogram buckets. Here's how you do that in Halide:
        histogram = hl.Func("histogram")

        # Histogram buckets start as zero.
        histogram[x] = 0

        # Define a multi-dimensional reduction domain over the input image:
        r = hl.RDom([(0, input.width()), (0, input.height())])

        # For every point in the reduction domain, increment the
        # histogram bucket corresponding to the intensity of the
        # input image at that point.
        histogram[input[r.x, r.y]] += 1

        halide_result = histogram.realize([256])

        # The equivalent Python is:
        py_result = np.empty((256), dtype=np.int32)
        for xx in range(256):
            py_result[xx] = 0

        for r_y in range(input.height()):
            for r_x in range(input.width()):
                py_result[input_data[r_y, r_x]] += 1

        # Check the answers agree:
        for xx in range(256):
            assert py_result[xx] == halide_result[xx], \
                "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

    # Scheduling update steps
    if True:
        # The pure variables in an update step and can be
        # parallelized, vectorized, split, etc as usual.

        # Vectorizing, splitting, or parallelize the variables that
        # are part of the reduction domain is trickier. We'll cover
        # that in a later lesson.

        # Consider the definition:
        f = hl.Func("x")
        f[x, y] = x * y
        # Set the second row to equal the first row.
        f[x, 1] = f[x, 0]
        # Set the second column to equal the first column plus 2.
        f[1, y] = f[0, y] + 2

        # The pure variables in each stage can be scheduled
        # independently. To control the pure definition, we schedule
        # as we have done in the past. The following code vectorizes
        # and parallelizes the pure definition only.
        f.vectorize(x, 4).parallel(y)

        # We use hl.Func::update(int) to get a handle to an update step
        # for the purposes of scheduling. The following line
        # vectorizes the first update step across x. We can't do
        # anything with y for this update step, because it doesn't
        # use y.
        f.update(0).vectorize(x, 4)

        # Now we parallelize the second update step in chunks of size
        # 4.
        yo, yi = hl.Var("yo"), hl.Var("yi")
        f.update(1).split(y, yo, yi, 4).parallel(yo)

        halide_result = f.realize([16, 16])

        # Here's the equivalent (serial) C:
        py_result = np.empty((16, 16), dtype=np.int32)

        # Pure step. Vectorized in x and parallelized in y.
        for yy in range(16):  # Should be a parallel for loop
            for x_vec in range(4):
                xx = [x_vec * 4, x_vec * 4 + 1, x_vec * 4 + 2, x_vec * 4 + 3]
                py_result[yy][xx[0]] = xx[0] * yy
                py_result[yy][xx[1]] = xx[1] * yy
                py_result[yy][xx[2]] = xx[2] * yy
                py_result[yy][xx[3]] = xx[3] * yy

        # First update. Vectorized in x.
        for x_vec in range(4):
            xx = [x_vec * 4, x_vec * 4 + 1, x_vec * 4 + 2, x_vec * 4 + 3]
            py_result[1][xx[0]] = py_result[0][xx[0]]
            py_result[1][xx[1]] = py_result[0][xx[1]]
            py_result[1][xx[2]] = py_result[0][xx[2]]
            py_result[1][xx[3]] = py_result[0][xx[3]]

        # Second update. Parallelized in chunks of size 4 in y.
        for yo in range(4):  # Should be a parallel for loop
            for yi in range(4):
                yy = yo * 4 + yi
                py_result[yy][1] = py_result[yy][0] + 2

        # Check the C and Halide results match:
        for yy in range(16):
            for xx in range(16):
                assert halide_result[xx, yy] == py_result[yy][xx], \
                    "halide_result(%d, %d) = %d instead of %d" % (
                        xx, yy, halide_result[xx, yy], py_result[yy][xx])

    # That covers how to schedule the variables within a hl.Func that
    # uses update steps, but what about producer-consumer
    # relationships that involve compute_at and store_at? Let's
    # examine a reduction as a producer, in a producer-consumer pair.
    if True:
        # Because an update does multiple passes over a stored array,
        # it's not meaningful to inline them. So the default schedule
        # for them does the closest thing possible. It computes them
        # in the innermost loop of their consumer. Consider this
        # trivial example:
        producer, consumer = hl.Func("producer"), hl.Func("consumer")
        producer[x] = x * 17
        producer[x] += 1
        consumer[x] = 2 * producer[x]
        halide_result = consumer.realize([10])

        # The equivalent Python is:
        py_result = np.empty((10), dtype=np.int32)
        for xx in range(10):
            producer_storage = np.empty((1), dtype=np.int32)
            # Pure step for producer
            producer_storage[0] = xx * 17
            # Update step for producer
            producer_storage[0] = producer_storage[0] + 1
            # Pure step for consumer
            py_result[xx] = 2 * producer_storage[0]

        # Check the results match
        for xx in range(10):
            assert halide_result[xx] == py_result[xx], \
                "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

        # For all other compute_at/store_at options, the reduction
        # gets placed where you would expect, somewhere in the loop
        # nest of the consumer.

    # Now let's consider a reduction as a consumer in a
    # producer-consumer pair. This is a little more involved.
    if True:
        if True:
            # Case 1: The consumer references the producer in the pure step
            # only.
            producer, consumer = hl.Func("producer"), hl.Func("consumer")
            # The producer is pure.
            producer[x] = x * 17
            consumer[x] = 2 * producer[x]
            consumer[x] += 1

            # The valid schedules for the producer in this case are
            # the default schedule - inlined, and also:
            #
            # 1) producer.compute_at(x), which places the computation of
            # the producer inside the loop over x in the pure step of the
            # consumer.
            #
            # 2) producer.compute_root(), which computes all of the
            # producer ahead of time.
            #
            # 3) producer.store_root().compute_at(x), which allocates
            # space for the consumer outside the loop over x, but fills
            # it in as needed inside the loop.
            #
            # Let's use option 1.

            producer.compute_at(consumer, x)

            halide_result = consumer.realize([10])

            # The equivalent Python is:
            py_result = np.empty((10), dtype=np.int32)

            # Pure step for the consumer
            for xx in range(10):
                # Pure step for producer
                producer_storage = np.empty((1), dtype=np.int32)
                producer_storage[0] = xx * 17
                py_result[xx] = 2 * producer_storage[0]

            # Update step for the consumer
            for xx in range(10):
                py_result[xx] += 1

            # All of the pure step is evaluated before any of the
            # update step, so there are two separate loops over x.

            # Check the results match
            for xx in range(10):
                assert halide_result[xx] == py_result[xx], \
                    "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

        if True:
            # Case 2: The consumer references the producer in the update step
            # only
            producer, consumer = hl.Func("producer"), hl.Func("consumer")
            producer[x] = x * 17
            consumer[x] = x
            consumer[x] += producer[x]

            # Again we compute the producer per x coordinate of the
            # consumer. This places producer code inside the update
            # step of the producer, because that's the only step that
            # uses the producer.
            producer.compute_at(consumer, x)

            # Note however, that we didn't say:
            #
            # producer.compute_at(consumer.update(0), x).
            #
            # Scheduling is done with respect to Vars of a hl.Func, and
            # the Vars of a hl.Func are shared across the pure and
            # update steps.

            halide_result = consumer.realize([10])

            # The equivalent Python is:
            py_result = np.empty((10), dtype=np.int32)
            # Pure step for the consumer
            for xx in range(10):
                py_result[xx] = xx

            # Update step for the consumer
            for xx in range(10):
                # Pure step for producer
                producer_storage = np.empty((1), dtype=np.int32)
                producer_storage[0] = xx * 17
                py_result[xx] += producer_storage[0]

            # Check the results match
            for xx in range(10):
                assert halide_result[xx] == py_result[xx], \
                    "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

        if True:
            # Case 3: The consumer references the producer in
            # multiple steps that share common variables
            producer, consumer = hl.Func("producer"), hl.Func("consumer")
            producer[x] = x * 17
            consumer[x] = producer[x] * x
            consumer[x] += producer[x]

            # Again we compute the producer per x coordinate of the
            # consumer. This places producer code inside both the
            # pure and the update step of the producer. So there ends
            # up being two separate realizations of the producer, and
            # redundant work occurs.
            producer.compute_at(consumer, x)

            halide_result = consumer.realize([10])

            # The equivalent Python is:
            py_result = np.empty((10), dtype=np.int32)
            # Pure step for the consumer
            for xx in range(10):
                # Pure step for producer
                producer_storage = np.empty((1), dtype=np.int32)
                producer_storage[0] = xx * 17
                py_result[xx] = producer_storage[0] * xx

            # Update step for the consumer
            for xx in range(10):
                # Another copy of the pure step for producer
                producer_storage = np.empty((1), dtype=np.int32)
                producer_storage[0] = xx * 17
                py_result[xx] += producer_storage[0]

            # Check the results match
            for xx in range(10):
                assert halide_result[xx] == py_result[xx], \
                    "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

        if True:
            # Case 4: The consumer references the producer in
            # multiple steps that do not share common variables
            producer, consumer = hl.Func("producer"), hl.Func("consumer")
            producer[x, y] = x * y
            consumer[x, y] = x + y
            consumer[x, 0] = producer[x, x - 1]
            consumer[0, y] = producer[y, y - 1]

            # In this case neither producer.compute_at(consumer, x)
            # nor producer.compute_at(consumer, y) will work, because
            # either one fails to cover one of the uses of the
            # producer. So we'd have to inline producer, or use
            # producer.compute_root().

            # Let's say we really really want producer to be
            # compute_at the inner loops of both consumer update
            # steps. Halide doesn't allow multiple different
            # schedules for a single hl.Func, but we can work around it
            # by making two wrappers around producer, and scheduling
            # those instead:

            # Attempt 2:
            producer_wrapper_1, producer_wrapper_2, consumer_2 = hl.Func(), hl.Func(), hl.Func()
            producer_wrapper_1[x, y] = producer[x, y]
            producer_wrapper_2[x, y] = producer[x, y]

            consumer_2[x, y] = x + y
            consumer_2[x, 0] += producer_wrapper_1[x, x - 1]
            consumer_2[0, y] += producer_wrapper_2[y, y - 1]

            # The wrapper functions give us two separate handles on
            # the producer, so we can schedule them differently.
            producer_wrapper_1.compute_at(consumer_2, x)
            producer_wrapper_2.compute_at(consumer_2, y)

            halide_result = consumer_2.realize([10, 10])

            # The equivalent Python is:
            py_result = np.empty((10, 10), dtype=np.int32)

            # Pure step for the consumer
            for yy in range(10):
                for xx in range(10):
                    py_result[yy][xx] = xx + yy

            # First update step for consumer
            for xx in range(10):
                producer_wrapper_1_storage = np.empty((1), dtype=np.int32)
                producer_wrapper_1_storage[0] = xx * (xx - 1)
                py_result[0][xx] += producer_wrapper_1_storage[0]

            # Second update step for consumer
            for yy in range(10):
                producer_wrapper_2_storage = np.empty((1), dtype=np.int32)
                producer_wrapper_2_storage[0] = yy * (yy - 1)
                py_result[yy][0] += producer_wrapper_2_storage[0]

            # Check the results match
            for yy in range(10):
                for xx in range(10):
                    assert halide_result[xx, yy] == py_result[yy][xx], \
                        "halide_result(%d, %d) = %d instead of %d" % (
                            xx, yy, halide_result[xx, yy], py_result[yy][xx])

        if True:
            # Case 5: Scheduling a producer under a reduction domain
            # variable of the consumer.

            # We are not just restricted to scheduling producers at
            # the loops over the pure variables of the consumer. If a
            # producer is only used within a loop over a reduction
            # domain (hl.RDom) variable, we can also schedule the
            # producer there.

            producer, consumer = hl.Func("producer"), hl.Func("consumer")

            r = hl.RDom([(0, 5)])
            producer[x] = x * 17
            consumer[x] = x + 10
            consumer[x] += r + producer[x + r]

            producer.compute_at(consumer, r)

            halide_result = consumer.realize([10])

            # The equivalent Python is:
            py_result = np.empty((10), dtype=np.int32)
            # Pure step for the consumer.
            for xx in range(10):
                py_result[xx] = xx + 10

            # Update step for the consumer.
            for xx in range(10):
                # The loop over the reduction domain is always the inner loop.
                for rr in range(5):
                    # We've schedule the storage and computation of
                    # the producer here. We just need a single value.
                    producer_storage = np.empty((1), dtype=np.int32)
                    # Pure step of the producer.
                    producer_storage[0] = (xx + rr) * 17

                    # Now use it in the update step of the consumer.
                    py_result[xx] += rr + producer_storage[0]

            # Check the results match
            for xx in range(10):
                assert halide_result[xx] == py_result[xx], \
                    "halide_result(%d) = %d instead of %d" % (xx, halide_result[xx], py_result[xx])

    # A real-world example of a reduction inside a producer-consumer chain.
    if True:
        # The default schedule for a reduction is a good one for
        # convolution-like operations. For example, the following
        # computes a 5x5 box-blur of our grayscale test image with a
        # hl.clamp-to-edge boundary condition:

        # First add the boundary condition.
        clamped = hl.BoundaryConditions.repeat_edge(input)

        # Define a 5x5 box that starts at (-2, -2)
        r = hl.RDom([(-2, 5), (-2, 5)])

        # Compute the 5x5 sum around each pixel.
        local_sum = hl.Func("local_sum")
        local_sum[x, y] = 0  # Compute the sum as a 32-bit integer
        local_sum[x, y] += clamped[x + r.x, y + r.y]

        # Divide the sum by 25 to make it an average
        blurry = hl.Func("blurry")
        blurry[x, y] = hl.cast(hl.UInt(8), local_sum[x, y] / 25)

        halide_result = blurry.realize([input.width(), input.height()])

        # The default schedule will inline 'clamped' into the update
        # step of 'local_sum', because clamped only has a pure
        # definition, and so its default schedule is fully-inlined.
        # We will then compute local_sum per x coordinate of blurry,
        # because the default schedule for reductions is
        # compute-innermost. Here's the equivalent Python:

        #cast_to_uint8 = lambda x_: np.array([x_], dtype=np.uint8)[0]
        local_sum = np.empty((1), dtype=np.int32)

        py_result = hl.Buffer(hl.UInt(8), [input.width(), input.height()])
        for yy in range(input.height()):
            for xx in range(input.width()):
                # FIXME this loop is quite slow
                # Pure step of local_sum
                local_sum[0] = 0
                # Update step of local_sum
                for r_y in range(-2, 2 + 1):
                    for r_x in range(-2, 2 + 1):
                        # The clamping has been inlined into the update step.
                        clamped_x = min(max(xx + r_x, 0), input.width() - 1)
                        clamped_y = min(max(yy + r_y, 0), input.height() - 1)
                        local_sum[0] += input[clamped_x, clamped_y]

                # Pure step of blurry
                # py_result(x, y) = (uint8_t)(local_sum[0] / 25)
                #py_result[xx, yy] = cast_to_uint8(local_sum[0] / 25)
                # hl.cast done internally
                py_result[xx, yy] = int(local_sum[0] / 25)

        # Check the results match
        for yy in range(input.height()):
            for xx in range(input.width()):
                assert halide_result[xx, yy] == py_result[xx, yy], \
                    "halide_result(%d, %d) = %d instead of %d" % (
                        xx, yy, halide_result[xx, yy], py_result[xx, yy])

    # Reduction helpers.
    if True:
        # There are several reduction helper functions provided in
        # Halide.h, which compute small reductions and schedule them
        # innermost into their consumer. The most useful one is
        # "sum".
        f1 = hl.Func("f1")
        r = hl.RDom([(0, 100)])
        f1[x] = hl.sum(r + x) * 7

        # Sum creates a small anonymous hl.Func to do the reduction. It's
        # equivalent to:
        f2, anon = hl.Func("f2"), hl.Func("anon")
        anon[x] = 0
        anon[x] += r + x
        f2[x] = anon[x] * 7

        # So even though f1 references a reduction domain, it is a
        # pure function. The reduction domain has been swallowed to
        # define the inner anonymous reduction.
        halide_result_1 = f1.realize([10])
        halide_result_2 = f2.realize([10])

        # The equivalent Python is:
        py_result = np.empty((10), dtype=np.int32)
        for xx in range(10):
            anon = np.empty((1), dtype=np.int32)
            anon[0] = 0
            for rr in range(100):
                anon[0] += rr + xx

            py_result[xx] = anon[0] * 7

        # Check they all match.
        for xx in range(10):
            assert halide_result_1[xx] == py_result[xx], \
                "halide_result_1(%d) = %d instead of %d" % (xx, halide_result_1[xx], py_result[xx])
            assert halide_result_2[xx] == py_result[xx], \
                "halide_result_2(%d) = %d instead of %d" % (xx, halide_result_2[xx], py_result[xx])

    print("Success!")
    return 0


if __name__ == "__main__":
    main()



try:
    import imageio.v2 as imageio
except:
    import imageio
import numpy as np


def is_interleaved(im):
    """If the given buffer is 3-dimensional and appears to have an interleaved
    layout, return True. Otherwise, return False."""

    # Assume that 'interleaved' will only apply to channels <= 4
    mv = memoryview(im)
    return mv.ndim == 3 and mv.strides[2] == 1 and mv.shape[2] in [1, 2, 3, 4]


def copy_to_interleaved(im):
    """If the given buffer is 3-dimensional and appears to be planar
    layout, return a copy that is in interleaved form. Otherwise, return
    an unchanged copy of the input. Note that this call will always return
    a copy, leaving the input unchanged."""
    mv = memoryview(im)
    if mv.ndim == 3 and not is_interleaved(mv):
        # We are presumably planar, in (c, y, x) order; we need (y, x, c) order
        mv = np.moveaxis(mv, 0, 2)
        mv = np.copy(mv, order="F")
        return mv
    else:
        return im


def copy_to_planar(im):
    """If the given buffer is 3-dimensional and appears to be interleaved
    layout, return a copy that is in planar form. Otherwise, return
    an unchanged copy of the input. Note that this call will always return
    a copy, leaving the input unchanged."""
    mv = memoryview(im)
    if is_interleaved(mv):
        # Interleaved will be in (y, x, c) order; we want (c, y, x) order
        # (which hl.Buffer will reverse into x, y, c order)
        mv = np.moveaxis(mv, 2, 0)
        mv = np.copy(mv, order="C")
        return mv
    else:
        return im


def imread(uri, format=None, **kwargs):
    """halide.imageio.imread is a thin wrapper around imagio.imread,
    except that for 3-dimensional images that appear to be interleaved,
    the result is converted to a planar layout before returning."""
    return copy_to_planar(imageio.imread(uri, format, **kwargs))


def imwrite(uri, im, format=None, **kwargs):
    """halide.imageio.imwrite is a thin wrapper around imagio.imwrite,
    except that for 3-dimensional images that appear to be planar,
    a temporary interleaved copy of the input is made, which is used for
    writing."""
    imageio.imwrite(uri, copy_to_interleaved(im), format, **kwargs)



"""Wrap our operator (and gradient) in autograd."""

# We need to import torch before loading the custom modules
import torch as th
import halide_ops as ops

# TODO(mgharbi): maybe find a way to wrap function and module directly in C++
# instead of generating the C++ wrapper on the fly?

def _dispatch(opname, optype=th.float32, cuda=False):
    """
    Helper function that matches an opname and type to the Halide backend.

    This is based on the naming convention we use in this example. Functions are
    named: <opname>[_cuda]_<optype>.

    Args:
      opname(str): name of the base Halide function.
      optype(torch.dtype): pytorch's tensor datatype.
      cuda(bool): whether the operator should use cuda.

    Returns:
      op: a python function wrapping the requested Halide operator.
    """

    assert type(opname) == str, "opname should be a string"
    assert type(optype) == th.dtype, "optype should be a tensor datatype (torch.dtype)"

    if cuda:
        opname += "_cuda"

    if optype == th.float32:
        opname += "_float32"
    elif optype == th.float64:
        opname += "_float64"
    else:
        raise ValueError("Optype %s not recognized %s" % optype)
    op = getattr(ops, opname)
    if not hasattr(ops, opname):
        raise ValueError("Module has no operator %s" % opname)
    return op

def _forward_common(ctx, input_a, input_b):
    tp = input_a.dtype
    cuda = input_a.is_cuda
    assert tp == input_b.dtype, "inputs should have the same type"
    assert cuda == input_b.is_cuda, "inputs should be on the same device (cpu/gpu)"

    ctx.save_for_backward(input_a, input_b)

    fn_ = _dispatch("add", optype=tp, cuda=cuda)

    # Create an output tensor with the proper dimensions
    out = input_a.new()
    out.resize_(input_a.shape)

    fn_(input_a, input_b, out)
    return out

def _backward_common(ctx, d_out, backward_op):
    tp = d_out.dtype
    cuda = d_out.is_cuda

    input_a = ctx.saved_tensors[0]
    input_b = ctx.saved_tensors[1]

    # Fetch the correct Halide operator for the type/device used
    fn_ = _dispatch(backward_op, optype=tp, cuda=cuda)

    d_input_a = d_out.new()
    d_input_b = d_out.new()
    d_input_a.resize_(d_out.shape)
    d_input_b.resize_(d_out.shape)

    fn_(input_a, input_b, d_out.contiguous(), d_input_a, d_input_b)
    return d_input_a, d_input_b

# TODO(srj): surely there's a better way to do this,
# but PyTorch seems to make it tricky to pass in
# extra info to the backward() method.
class AddFunction_Grad(th.autograd.Function):
  """Version using the manually-written backprop"""
  def __init__(self):
      super(AddFunction_Grad, self).__init__()

  @staticmethod
  def forward(ctx, input_a, input_b):
      return _forward_common(ctx, input_a, input_b)

  @staticmethod
  def backward(ctx, d_out):
      return _backward_common(ctx, d_out, "add_grad")

class AddFunction_HalideGrad(th.autograd.Function):
  """Version using the Halide-generated backprop"""
  def __init__(self):
      super(AddFunction_HalideGrad, self).__init__()

  @staticmethod
  def forward(ctx, input_a, input_b):
      return _forward_common(ctx, input_a, input_b)

  @staticmethod
  def backward(ctx, d_out):
      return _backward_common(ctx, d_out, "add_halidegrad")

class Add(th.nn.Module):
    """Defines a module that uses our autograd function.

    This is so we can use it as an operator.
    """
    def __init__(self, backward_op):
        super(Add, self).__init__()
        if backward_op == "add_grad":
          self._adder = AddFunction_Grad
        elif backward_op == "add_halidegrad":
          self._adder = AddFunction_HalideGrad
        else:
          assert False

    def forward(self, a, b):
        return self._adder.apply(a, b)



"""
Fast image interpolation using a pyramid.
"""

import halide as hl


def _func_list(name, size):
    """Return a list containing `size` Funcs, named `name_n` for n in 0..size-1."""
    return [hl.Func("%s_%d" % (name, i)) for i in range(size)]


@hl.alias(
    interpolate_Mullapudi2016={"autoscheduler": "Mullapudi2016"},
)
@hl.generator()
class interpolate:
    levels = hl.GeneratorParam(10)

    input_buf = hl.InputBuffer(hl.Float(32), 3)
    output_buf = hl.OutputBuffer(hl.Float(32), 3)

    def generate(self):
        g = self

        x, y, c = hl.vars("x y c")

        # Input must have four color channels - rgba
        g.input_buf.dim(2).set_bounds(0, 4)

        downsampled = _func_list("downsampled", g.levels)
        downx = _func_list("downx", g.levels)
        interpolated = _func_list("interpolated", g.levels)
        upsampled = _func_list("upsampled", g.levels)
        upsampledx = _func_list("upsampledx", g.levels)

        clamped = hl.BoundaryConditions.repeat_edge(g.input_buf)

        downsampled[0][x, y, c] = hl.select(
            c < 3,
            clamped[x, y, c] * clamped[x, y, 3],
            clamped[x, y, 3],
        )

        for l in range(1, g.levels):
            prev = downsampled[l - 1]

            if l == 4:
                # Also add a boundary condition at a middle pyramid level
                # to prevent the footprint of the downsamplings to extend
                # too far off the base image. Otherwise we look 512
                # pixels off each edge.
                w = g.input_buf.width() / (1 << (l - 1))
                h = g.input_buf.height() / (1 << (l - 1))
                prev = hl.lambda_func(
                    x, y, c, prev[hl.clamp(x, 0, w), hl.clamp(y, 0, h), c]
                )

            downx[l][x, y, c] = (
                prev[x * 2 - 1, y, c] + 2 * prev[x * 2, y, c] + prev[x * 2 + 1, y, c]
            ) * 0.25

            downsampled[l][x, y, c] = (
                downx[l][x, y * 2 - 1, c]
                + 2 * downx[l][x, y * 2, c]
                + downx[l][x, y * 2 + 1, c]
            ) * 0.25

        interpolated[g.levels - 1][x, y, c] = downsampled[g.levels - 1][x, y, c]

        for l in range(g.levels - 2, -1, -1):
            upsampledx[l][x, y, c] = (
                interpolated[l + 1][x / 2, y, c]
                + interpolated[l + 1][(x + 1) / 2, y, c]
            ) / 2
            upsampled[l][x, y, c] = (
                upsampledx[l][x, y / 2, c] + upsampledx[l][x, (y + 1) / 2, c]
            ) / 2
            alpha = 1.0 - downsampled[l][x, y, 3]
            interpolated[l][x, y, c] = (
                downsampled[l][x, y, c] + alpha * upsampled[l][x, y, c]
            )

        g.output_buf[x, y, c] = interpolated[0][x, y, c] / interpolated[0][x, y, 3]

        # Schedule
        if g.using_autoscheduler():
            # nothing
            pass
        elif g.target().has_gpu_feature():
            # 0.86ms on a 2060 RTX
            yo, yi, xo, xi, ci, xii, yii = hl.vars("yo yi xo xi ci xii yii")

            (
                g.output_buf.bound(x, 0, g.input_buf.width())
                .bound(y, 0, g.input_buf.height())
                .bound(c, 0, 3)
                .reorder(c, x, y)
                .tile(x, y, xi, yi, 32, 32, hl.TailStrategy.RoundUp)
                .tile(xi, yi, xii, yii, 2, 2)
                .gpu_blocks(x, y)
                .gpu_threads(xi, yi)
                .unroll(xii)
                .unroll(yii)
                .unroll(c)
            )

            for l in range(1, g.levels):
                (
                    downsampled[l]
                    .compute_root()
                    .reorder(c, x, y)
                    .unroll(c)
                    .gpu_tile(x, y, xi, yi, 16, 16)
                )

            for l in range(3, g.levels, 2):
                (
                    interpolated[l]
                    .compute_root()
                    .reorder(c, x, y)
                    .tile(x, y, xi, yi, 32, 32, hl.TailStrategy.RoundUp)
                    .tile(xi, yi, xii, yii, 2, 2)
                    .gpu_blocks(x, y)
                    .gpu_threads(xi, yi)
                    .unroll(xii)
                    .unroll(yii)
                    .unroll(c)
                )

            (
                upsampledx[1]
                .compute_at(g.output_buf, x)
                .reorder(c, x, y)
                .tile(x, y, xi, yi, 2, 1)
                .unroll(xi)
                .unroll(yi)
                .unroll(c)
                .gpu_threads(x, y)
            )

            (
                interpolated[1]
                .compute_at(g.output_buf, x)
                .reorder(c, x, y)
                .tile(x, y, xi, yi, 2, 2)
                .unroll(xi)
                .unroll(yi)
                .unroll(c)
                .gpu_threads(x, y)
            )

            (
                interpolated[2]
                .compute_at(g.output_buf, x)
                .reorder(c, x, y)
                .unroll(c)
                .gpu_threads(x, y)
            )

        else:
            # 4.54ms on an Intel i9-9960X using 16 threads
            xo, xi, yo, yi = hl.vars("xo xi yo yi")
            vec = g.natural_vector_size(hl.Float(32))
            for l in range(1, g.levels - 1):
                # We must refer to the downsampled stages in the
                # upsampling later, so they must all be
                # compute_root or redundantly recomputed, as in
                # the local_laplacian app.
                (
                    downsampled[l]
                    .compute_root()
                    .reorder(x, c, y)
                    .split(y, yo, yi, 8)
                    .parallel(yo)
                    .vectorize(x, vec)
                )

            # downsampled[0] takes too long to compute_root, so
            # we'll redundantly recompute it instead.  Make a
            # separate clone of it in the first downsampled stage
            # so that we can schedule the two versions
            # separately.
            (
                downsampled[0]
                .clone_in(downx[1])
                .store_at(downsampled[1], yo)
                .compute_at(downsampled[1], yi)
                .reorder(c, x, y)
                .unroll(c)
                .vectorize(x, vec)
            )

            (
                g.output_buf.bound(x, 0, g.input_buf.width())
                .bound(y, 0, g.input_buf.height())
                .bound(c, 0, 3)
                .split(x, xo, xi, vec)
                .split(y, yo, yi, 32)
                .reorder(xi, c, xo, yi, yo)
                .unroll(c)
                .vectorize(xi)
                .parallel(yo)
            )

            for l in range(1, g.levels):
                (
                    interpolated[l]
                    .store_at(g.output_buf, yo)
                    .compute_at(g.output_buf, yi)
                    .vectorize(x, vec)
                )

        # Estimates (for autoscheduler; ignored otherwise)
        (
            g.input_buf.dim(0)
            .set_estimate(0, 1536)
            .dim(1)
            .set_estimate(0, 2560)
            .dim(2)
            .set_estimate(0, 4)
        )
        (
            g.output_buf.output_buffer()
            .dim(0)
            .set_estimate(0, 1536)
            .dim(1)
            .set_estimate(0, 2560)
            .dim(2)
            .set_estimate(0, 3)
        )


if __name__ == "__main__":
    hl.main()



from bitcpp import bitcpp
from bitpy import bitpy
import numpy as np


def test(fn):
    # Note that in Halide, Buffer<bool> and Buffer<uint8> have identical memory
    # layout -- a bool takes an entire byte in a buffer -- but distinct types,
    # so we must construct a Buffer here that has that correct type. (If we
    # passed in a uint8 buffer, we'd fail with 'wrong type' exception.)
    #
    # Python's `array` module doesn't support boolean entries, but libraries
    # that support Buffer Protocol (eg numpy.ndarray) do, so we'll use one of those.
    input_bools = np.ndarray([4], dtype=bool)
    output_bools = np.ndarray([4], dtype=bool)

    for i in range(0, 4):
        input_bools[i] = (i & 1) != 0

    fn(input_bools, True, output_bools)
    for i in range(0, 4):
        assert output_bools[i] == True

    fn(input_bools, False, output_bools)
    for i in range(0, 4):
        assert output_bools[i] == ((i & 1) != 0)


if __name__ == "__main__":
    test(bitcpp)
    test(bitpy)



import model_cpp


class Model():
    def __init__(self):
        self.pipeline = None

    def BuildFromOnnxModel(self, onnx_model, expected_dim_sizes=None,
                           layout=model_cpp.Layout.NumPy):
        assert onnx_model
        if not expected_dim_sizes:
            expected_dim_sizes = {}

        if type(onnx_model) is str:
            onnx_model = onnx_model.encode()
            self.pipeline = model_cpp.ConvertOnnxModel(onnx_model,
                expected_dim_sizes, layout)
        elif type(onnx_model) is bytes:
            self.pipeline = model_cpp.ConvertOnnxModel(onnx_model,
                expected_dim_sizes, layout)
        else:
            # protobuf don't support swig, so we have to convert them to string.
            model_str = onnx_model.SerializeToString()
            self.pipeline = model_cpp.ConvertOnnxModel(model_str,
                expected_dim_sizes, layout)

    def OptimizeSchedule(self):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        return model_cpp.AutoSchedule(self.pipeline)

    def run(self, inputs, device=''):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        return model_cpp.Run(self.pipeline, inputs, device)

    def Benchmark(self, num_iters=5, device=''):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        return model_cpp.Benchmark(self.pipeline, num_iters, device)

    def Compile(self, func_name, lib_name):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        return model_cpp.Compile(self.pipeline, func_name, lib_name)

    def PrintLoopNest(self):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        model_cpp.PrintLoopNest(self.pipeline)

    def PrintLoweredStatement(self):
        if not self.pipeline:
            raise Exception("model not initialized, call BuildFromOnnxModel first")
        model_cpp.PrintLoweredStatement(self.pipeline)



def make_oklch(l, c, h):
    return ("oklch(%.1f%% %.2f %.0f)" % (l * 100, c, h))


STEPS = 20

for i in range(STEPS):
    f = i / (STEPS - 1)
    col = make_oklch(0.9 - f * 0.5, 0.05 + 0.1 * f, 140)
    print(".block-CostColor%d:first-child { border-left: 8px solid %s; }" % (i, col))
print(".block-CostColorNone:first-child { border-left: transparent; }")
print()

for i in range(STEPS):
    f = i / (STEPS - 1)
    col = make_oklch(0.9 - f * 0.5, 0.05 + 0.1 * f, 140)
    print(".line-CostColor%d:first-child { border-right: 8px solid %s; }" % (i, col))
print(".line-CostColorNone:first-child { border-right: transparent; }")
print()


for i in range(STEPS):
    f = i / (STEPS - 1)
    col = make_oklch(0.9 - f * 0.5, 0.05 + 0.1 * f, 300)
    print(".block-CostColor%d:last-child { border-left: 8px solid %s; }" % (i, col))
print(".block-CostColorNone:last-child { border-left: transparent; }")
print()

for i in range(STEPS):
    f = i / (STEPS - 1)
    col = make_oklch(0.9 - f * 0.5, 0.05 + 0.1 * f, 300)
    print(".line-CostColor%d:last-child { border-right: 8px solid %s; }" % (i, col))
print(".line-CostColorNone:last-child { border-right: transparent; }")
print()
